# Copyright 2016-2021 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tf/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Test the featurenet models."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import parameterized

from tensorflow.python.framework import test_util as tf_test_util
from tensorflow.python.keras import keras_parameterized

from tensorflow.keras import backend as K

from deepcell.model_zoo import featurenet


class FeatureNetTest(keras_parameterized.TestCase):

    @keras_parameterized.run_all_keras_modes
    @parameterized.named_parameters([
        {
            'testcase_name': 'dilated_include_top',
            'include_top': True,
            'dilated': True,
            'padding_mode': 'reflect',
            'padding': False,
            'multires': False,
            'location': False,
            'shape': (32, 32, 1),
            'data_format': 'channels_last'
        },
        {
            'testcase_name': 'dilated_include_top_cf',
            'include_top': True,
            'dilated': True,
            'padding_mode': 'reflect',
            'padding': False,
            'multires': False,
            'location': False,
            'shape': (32, 32, 1),
            'data_format': 'channels_first'
        },
        # {
        #     'testcase_name': 'not_dilated_include_top',
        #     'include_top': True,
        #     'dilated': False,
        #     'padding_mode': 'reflect',
        #     'padding': True,
        #     'multires': False,
        #     'location': False,
        #     'shape': (32, 32, 1),
        #     'data_format': 'channels_last',
        # },
        # {
        #     'testcase_name': 'not_dilated_include_top_cf',
        #     'include_top': True,
        #     'dilated': False,
        #     'padding_mode': 'reflect',
        #     'padding': True,
        #     'multires': False,
        #     'location': False,
        #     'shape': (32, 32, 1),
        #     'data_format': 'channels_first',
        # },
        {
            'testcase_name': 'dilated_no_top',
            'include_top': False,
            'dilated': True,
            'padding_mode': 'zero',
            'padding': False,
            'multires': False,
            'location': False,
            'shape': (33, 33, 1),
            'data_format': 'channels_last',
        },
        {
            'testcase_name': 'dilated_no_top_cf',
            'include_top': False,
            'dilated': True,
            'padding_mode': 'zero',
            'padding': False,
            'multires': False,
            'location': False,
            'shape': (33, 33, 1),
            'data_format': 'channels_first',
        },
        {
            'testcase_name': 'not_dilated_no_top',
            'include_top': False,
            'dilated': False,
            'padding_mode': 'reflect',
            'padding': False,
            'multires': False,
            'location': False,
            'shape': (32, 32, 1),
            'data_format': 'channels_last',
        },
        {
            'testcase_name': 'not_dilated_no_top_cf',
            'include_top': False,
            'dilated': False,
            'padding_mode': 'reflect',
            'padding': False,
            'multires': False,
            'location': False,
            'shape': (32, 32, 1),
            'data_format': 'channels_first',
        },
        {
            'testcase_name': 'not_dilated_no_top_location',
            'include_top': False,
            'dilated': False,
            'padding_mode': 'reflect',
            'padding': False,
            'multires': True,
            'location': True,
            'shape': (32, 32, 1),
            'data_format': 'channels_last',
        },
        {
            'testcase_name': 'not_dilated_no_top_location_cf',
            'include_top': False,
            'dilated': False,
            'padding_mode': 'reflect',
            'padding': False,
            'multires': True,
            'location': True,
            'shape': (32, 32, 1),
            'data_format': 'channels_first',
        },
        {
            'testcase_name': 'not_dilated_no_top_multires',
            'include_top': False,
            'dilated': False,
            'padding_mode': 'reflect',
            'padding': False,
            'multires': True,
            'location': False,
            'shape': (32, 32, 1),
            'data_format': 'channels_last',
        },
        {
            'testcase_name': 'not_dilated_no_top_multires_cf',
            'include_top': False,
            'dilated': False,
            'padding_mode': 'reflect',
            'padding': False,
            'multires': True,
            'location': False,
            'shape': (32, 32, 1),
            'data_format': 'channels_first',
        },
    ])
    def test_bn_feature_net_2D(self, include_top, padding, padding_mode, shape,
                               dilated, multires, location, data_format):
        n_features = 3
        n_dense_filters = 200

        # BAD: dilated=True, include_top=False
        # BAD: inputs != None

        with self.cached_session():
            K.set_image_data_format(data_format)
            model = featurenet.bn_feature_net_2D(
                include_top=include_top,
                dilated=dilated,
                input_shape=shape,
                n_features=n_features,
                n_dense_filters=n_dense_filters,
                padding=padding,
                padding_mode=padding_mode,
                multires=multires,
                VGG_mode=multires,
                location=location)
            self.assertEqual(len(model.output_shape), 4)
            output = n_features if include_top else n_dense_filters
            axis = 1 if data_format == 'channels_first' else -1
            self.assertEqual(model.output_shape[axis], output)

    @keras_parameterized.run_all_keras_modes
    @parameterized.named_parameters(
        *tf_test_util.generate_combinations_with_testcase_name(
            data_format=['channels_first', 'channels_last']))
    def test_bn_feature_net_2D_skip(self, data_format):
        receptive_field = 61
        n_features = 3
        n_dense_filters = 300
        input_shape = (256, 256, 1)
        n_skips = 1

        with self.cached_session():
            K.set_image_data_format(data_format)
            axis = 1 if data_format == 'channels_first' else -1

            fgbg_model = featurenet.bn_feature_net_skip_2D(
                receptive_field=receptive_field,
                input_shape=input_shape,
                n_features=n_features,
                n_dense_filters=n_dense_filters,
                n_skips=n_skips,
                last_only=False)

            self.assertIsInstance(fgbg_model.output, list)
            self.assertEqual(len(fgbg_model.output), n_skips + 1)

            model = featurenet.bn_feature_net_skip_2D(
                receptive_field=receptive_field,
                input_shape=input_shape,
                fgbg_model=fgbg_model,
                n_features=n_features,
                n_dense_filters=n_dense_filters,
                n_skips=n_skips,
                last_only=True)

            self.assertEqual(len(model.output_shape), 4)
            self.assertEqual(model.output_shape[axis], n_features)

    @keras_parameterized.run_all_keras_modes
    @parameterized.named_parameters([
        {
            'testcase_name': 'dilated_include_top',
            'include_top': True,
            'dilated': True,
            'padding_mode': 'reflect',
            'padding': False,
            'multires': False,
            'location': False,
            'shape': (10, 33, 33, 1),
            'data_format': 'channels_last',
            'temporal': None,
            'residual': False,
            'temporal_kernel_size': 3,
        },
        {
            'testcase_name': 'dilated_include_top_cf',
            'include_top': True,
            'dilated': True,
            'padding_mode': 'reflect',
            'padding': False,
            'multires': False,
            'location': False,
            'shape': (10, 33, 33, 1),
            'data_format': 'channels_first',
            'temporal': None,
            'residual': False,
            'temporal_kernel_size': 3,
        },
        # {
        #     'testcase_name': 'not_dilated_include_top',
        #     'include_top': True,
        #     'dilated': False,
        #     'padding_mode': 'reflect',
        #     'padding': True,
        #     'multires': False,
        #     'location': False,
        #     'data_format': 'channels_last',
        # },
        # {
        #     'testcase_name': 'not_dilated_include_top_cf',
        #     'include_top': True,
        #     'dilated': False,
        #     'padding_mode': 'reflect',
        #     'padding': True,
        #     'multires': False,
        #     'location': False,
        #     'data_format': 'channels_first',
        # },
        {
            'testcase_name': 'dilated_no_top',
            'include_top': False,
            'dilated': True,
            'padding_mode': 'zero',
            'padding': False,
            'multires': False,
            'location': False,
            'shape': (10, 32, 32, 1),
            'data_format': 'channels_last',
            'temporal': None,
            'residual': False,
            'temporal_kernel_size': 3,
        },
        {
            'testcase_name': 'dilated_no_top_cf',
            'include_top': False,
            'dilated': True,
            'padding_mode': 'zero',
            'padding': False,
            'multires': False,
            'location': False,
            'shape': (10, 32, 32, 1),
            'data_format': 'channels_first',
            'temporal': None,
            'residual': False,
            'temporal_kernel_size': 3,
        },
        {
            'testcase_name': 'not_dilated_no_top',
            'include_top': False,
            'dilated': False,
            'padding_mode': 'reflect',
            'padding': False,
            'multires': False,
            'location': False,
            'shape': (10, 33, 33, 1),
            'data_format': 'channels_last',
            'temporal': None,
            'residual': False,
            'temporal_kernel_size': 3,
        },
        {
            'testcase_name': 'not_dilated_no_top_cf',
            'include_top': False,
            'dilated': False,
            'padding_mode': 'reflect',
            'padding': False,
            'multires': False,
            'location': False,
            'shape': (10, 33, 33, 1),
            'data_format': 'channels_first',
            'temporal': None,
            'residual': False,
            'temporal_kernel_size': 3,
        },
        {
            'testcase_name': 'not_dilated_no_top_location',
            'include_top': False,
            'dilated': False,
            'padding_mode': 'reflect',
            'padding': False,
            'multires': True,
            'location': True,
            'shape': (10, 32, 32, 1),
            'data_format': 'channels_last',
            'temporal': None,
            'residual': False,
            'temporal_kernel_size': 3,
        },
        {
            'testcase_name': 'not_dilated_no_top_location_cf',
            'include_top': False,
            'dilated': False,
            'padding_mode': 'reflect',
            'padding': False,
            'multires': True,
            'location': True,
            'shape': (10, 32, 32, 1),
            'data_format': 'channels_first',
            'temporal': None,
            'residual': False,
            'temporal_kernel_size': 3,
        },
        {
            'testcase_name': 'not_dilated_no_top_multires',
            'include_top': False,
            'dilated': False,
            'padding_mode': 'reflect',
            'padding': False,
            'multires': True,
            'location': False,
            'shape': (10, 33, 33, 1),
            'data_format': 'channels_last',
            'temporal': None,
            'residual': False,
            'temporal_kernel_size': 3,
        },
        {
            'testcase_name': 'not_dilated_no_top_multires_cf',
            'include_top': False,
            'dilated': False,
            'padding_mode': 'reflect',
            'padding': False,
            'multires': True,
            'location': False,
            'shape': (10, 33, 33, 1),
            'data_format': 'channels_first',
            'temporal': None,
            'residual': False,
            'temporal_kernel_size': 3,
        },
        {
            'testcase_name': 'dilated_no_top_conv3D_nonresidual',
            'include_top': False,
            'dilated': True,
            'padding_mode': 'zero',
            'padding': False,
            'multires': False,
            'location': False,
            'shape': (10, 32, 32, 1),
            'data_format': 'channels_last',
            'temporal': 'conv',
            'residual': False,
            'temporal_kernel_size': 3,
        },
        {
            'testcase_name': 'dilated_no_top_conv3D_residual',
            'include_top': False,
            'dilated': True,
            'padding_mode': 'zero',
            'padding': False,
            'multires': False,
            'location': False,
            'shape': (10, 32, 32, 1),
            'data_format': 'channels_last',
            'temporal': 'conv',
            'residual': True,
            'temporal_kernel_size': 3,
        },
        {
            'testcase_name': 'dilated_no_top_convGRU2D_nonresidual',
            'include_top': False,
            'dilated': True,
            'padding_mode': 'zero',
            'padding': False,
            'multires': False,
            'location': False,
            'shape': (10, 32, 32, 1),
            'data_format': 'channels_last',
            'temporal': 'gru',
            'residual': False,
            'temporal_kernel_size': 3,
        },
        {
            'testcase_name': 'dilated_no_top_convgru2D_nonresidual',
            'include_top': False,
            'dilated': True,
            'padding_mode': 'zero',
            'padding': False,
            'multires': False,
            'location': False,
            'shape': (10, 32, 32, 1),
            'data_format': 'channels_last',
            'temporal': 'GRU',
            'residual': False,
            'temporal_kernel_size': 3,
        },
        {
            'testcase_name': 'dilated_no_top_convGRU2D_residual',
            'include_top': False,
            'dilated': True,
            'padding_mode': 'zero',
            'padding': False,
            'multires': False,
            'location': False,
            'shape': (10, 32, 32, 1),
            'data_format': 'channels_last',
            'temporal': 'gru',
            'residual': True,
            'temporal_kernel_size': 3,
        },
        {
            'testcase_name': 'dilated_no_top_convlstm2D_nonresidual',
            'include_top': False,
            'dilated': True,
            'padding_mode': 'zero',
            'padding': False,
            'multires': False,
            'location': False,
            'shape': (10, 32, 32, 1),
            'data_format': 'channels_last',
            'temporal': 'lstm',
            'residual': False,
            'temporal_kernel_size': 3,
        },
        {
            'testcase_name': 'dilated_no_top_convLSTM2D_nonresidual',
            'include_top': False,
            'dilated': True,
            'padding_mode': 'zero',
            'padding': False,
            'multires': False,
            'location': False,
            'shape': (10, 32, 32, 1),
            'data_format': 'channels_last',
            'temporal': 'LSTM',
            'residual': False,
            'temporal_kernel_size': 3,
        },
        {
            'testcase_name': 'dilated_no_top_convLSTM2D_residual',
            'include_top': False,
            'dilated': True,
            'padding_mode': 'zero',
            'padding': False,
            'multires': False,
            'location': False,
            'shape': (10, 32, 32, 1),
            'data_format': 'channels_last',
            'temporal': 'lstm',
            'residual': True,
            'temporal_kernel_size': 3,
        },
    ])
    def test_bn_feature_net_3D(self, include_top, padding, padding_mode, shape,
                               dilated, multires, location, data_format,
                               temporal, residual, temporal_kernel_size):
        n_features = 3
        n_dense_filters = 200
        n_frames = 5
        # input_shape = (10, 32, 32, 1)

        with self.cached_session():
            K.set_image_data_format(data_format)
            model = featurenet.bn_feature_net_3D(
                include_top=include_top,
                dilated=dilated,
                n_frames=n_frames,
                input_shape=shape,
                n_features=n_features,
                n_dense_filters=n_dense_filters,
                padding=padding,
                padding_mode=padding_mode,
                multires=multires,
                VGG_mode=multires,
                location=location,
                temporal=temporal,
                residual=residual,
                temporal_kernel_size=temporal_kernel_size)
            self.assertEqual(len(model.output_shape), 5 if dilated else 2)
            channel_axis = 1 if data_format == 'channels_first' else -1
            self.assertEqual(model.output_shape[channel_axis], n_features)

    @keras_parameterized.run_all_keras_modes
    @parameterized.named_parameters(
        *tf_test_util.generate_combinations_with_testcase_name(
            data_format=['channels_first', 'channels_last']))
    def test_bn_feature_net_3D_skip(self, data_format):
        receptive_field = 61
        n_features = 3
        n_dense_filters = 300
        input_shape = (10, 32, 32, 1)
        n_skips = 1
        temporal = None
        residual = False
        temporal_kernel_size = 3

        with self.cached_session():
            K.set_image_data_format(data_format)
            axis = 1 if data_format == 'channels_first' else -1

            fgbg_model = featurenet.bn_feature_net_skip_3D(
                receptive_field=receptive_field,
                input_shape=input_shape,
                n_features=n_features,
                n_dense_filters=n_dense_filters,
                n_skips=n_skips,
                last_only=False,
                temporal=temporal,
                residual=residual,
                temporal_kernel_size=temporal_kernel_size)

            self.assertIsInstance(fgbg_model.output, list)
            self.assertEqual(len(fgbg_model.output), n_skips + 1)

            model = featurenet.bn_feature_net_skip_3D(
                receptive_field=receptive_field,
                input_shape=input_shape,
                fgbg_model=fgbg_model,
                n_features=n_features,
                n_dense_filters=n_dense_filters,
                n_skips=n_skips,
                last_only=True,
                temporal=temporal,
                residual=residual,
                temporal_kernel_size=temporal_kernel_size)

            self.assertEqual(len(model.output_shape), 5)
            self.assertEqual(model.output_shape[axis], n_features)



# Copyright 2016-2021 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tf/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities plotting data"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import os

import numpy as np
import matplotlib.pyplot as plt

from matplotlib import animation
from tensorflow.python.keras import backend as K
from skimage.exposure import rescale_intensity
from skimage.segmentation import find_boundaries


def get_js_video(images, batch=0, channel=0, cmap='jet',
                 vmin=0, vmax=0, interval=200, repeat_delay=1000):
    """Create a JavaScript video as HTML for visualizing 3D data as a movie.

    Args:
        images (numpy.array): images to display as video
        batch (int): batch number of images to plot
        channel (int): channel index to plot
        vmin (int): lower end of data range covered by colormap
        vmax (int): upper end of data range covered by colormap

    Returns:
        str: JS HTML to display video
    """
    fig = plt.figure()

    ims = []
    plot_kwargs = {
        'animated': True,
        'cmap': cmap,
    }

    if vmax == 0:
        vmax = images.max()

    # TODO: do these not work for other cmaps?
    if cmap == 'cubehelix' or cmap == 'jet':
        plot_kwargs['vmin'] = vmin
        plot_kwargs['vmax'] = vmax

    for i in range(images.shape[1]):
        im = plt.imshow(images[batch, i, :, :, channel], **plot_kwargs)
        ims.append([im])

    ani = animation.ArtistAnimation(fig, ims, interval=interval, repeat_delay=repeat_delay)
    plt.close()
    return ani.to_jshtml()


def cf(x_coord, y_coord, sample_image):
    """Format x and y coordinates for printing

    Args:
        x_coord (int): X coordinate
        y_coord (int): y coordinate
        sample_image (numpy.array): Sample image for numpy arrays

    Returns:
        str: formatted coordinates ``(x, y, z)``.
    """
    numrows, numcols = sample_image.shape
    col = int(x_coord + 0.5)
    row = int(y_coord + 0.5)
    if 0 <= col < numcols and 0 <= row < numrows:
        z_coord = sample_image[row, col]
        return 'x=%1.4f, y=%1.4f, z=%1.4f' % (x_coord, y_coord, z_coord)
    return 'x=%1.4f, y=%1.4f' % (x_coord, y_coord)


def plot_training_data_2d(X, y, max_plotted=5):
    data_format = K.image_data_format()
    if max_plotted > y.shape[0]:
        max_plotted = y.shape[0]

    label_axis = 1 if K.image_data_format() == 'channels_first' else -1

    fig, ax = plt.subplots(max_plotted, y.shape[label_axis] + 1, squeeze=False)

    for i in range(max_plotted):
        X_i = X[i, 0, :, :] if data_format == 'channels_first' else X[i, :, :, 0]
        ax[i, 0].imshow(X_i, cmap=plt.get_cmap('gray'), interpolation='nearest')

        def form_coord(x_coord, y_coord):
            return cf(x_coord, y_coord, X_i)

        ax[i, 0].format_coord = form_coord
        ax[i, 0].axes.get_xaxis().set_visible(False)
        ax[i, 0].axes.get_yaxis().set_visible(False)

        for j in range(1, y.shape[label_axis] + 1):
            y_k = y[i, j - 1, :, :] if data_format == 'channels_first' else y[i, :, :, j - 1]
            ax[i, j].imshow(y_k, cmap=plt.get_cmap('gray'), interpolation='nearest')
            ax[i, j].axes.get_xaxis().set_visible(False)
            ax[i, j].axes.get_yaxis().set_visible(False)

    plt.show()


def plot_training_data_3d(X, y, num_image_stacks, frames_to_display=5):
    """Plot 3D training data

    Args:
        X (numpy.array): Raw 3D data
        y (numpy.array): Labels for 3D data
        num_image_stacks (int): number of independent 3D examples to plot
        frames_to_display (int): number of frames of X and y to display
    """
    data_format = K.image_data_format()
    fig, ax = plt.subplots(num_image_stacks, frames_to_display + 1, squeeze=False)

    for i in range(num_image_stacks):
        X_i = X[i, 0, :, :] if data_format == 'channels_first' else X[i, :, :, 0]
        ax[i, 0].imshow(X_i, cmap=plt.get_cmap('gray'), interpolation='nearest')

        def form_coord(x_coord, y_coord):
            return cf(x_coord, y_coord, X_i)

        ax[i, 0].format_coord = form_coord
        ax[i, 0].axes.get_xaxis().set_visible(False)
        ax[i, 0].axes.get_yaxis().set_visible(False)

        for j in range(frames_to_display):
            y_j = y[i, j, :, :] if data_format == 'channels_first' else y[i, :, :, j]
            ax[i, j + 1].imshow(y_j, cmap=plt.get_cmap('gray'), interpolation='nearest')
            ax[i, j + 1].axes.get_xaxis().set_visible(False)
            ax[i, j + 1].axes.get_yaxis().set_visible(False)
    plt.show()


def plot_error(loss_hist_file, saved_direc, plot_name):
    """Plot the training and validation error from the npz file

    Args:
        loss_hist_file (str): full path to .npz loss history file
        saved_direc (str): full path to directory where the plot is saved
        plot_name (str): the name of plot
    """
    loss_history = np.load(loss_hist_file)
    loss_history = loss_history['loss_history'][()]

    err = np.subtract(1, loss_history['acc'])
    val_err = np.subtract(1, loss_history['val_acc'])

    epoch = np.arange(1, len(err) + 1, 1)
    plt.plot(epoch, err)
    plt.plot(epoch, val_err)
    plt.title('Model Error')
    plt.xlabel('Epoch')
    plt.ylabel('Model Error')
    plt.legend(['Training error', 'Validation error'], loc='upper right')

    filename = os.path.join(saved_direc, plot_name)
    plt.savefig(filename, format='pdf')


def create_rgb_image(input_data, channel_colors):
    """Takes a stack of 1- or 2-channel data and converts it to an RGB image

    Args:
        input_data: 4D stack of images to be converted to RGB
        channel_colors: list specifying the color for each channel

    Returns:
        numpy.array: transformed version of input data into RGB version

    Raises:
        ValueError: if ``len(channel_colors)`` is not equal
            to number of channels
        ValueError: if invalid ``channel_colors`` provided
        ValueError: if input_data is not 4D, with 1 or 2 channels
    """

    if len(input_data.shape) != 4:
        raise ValueError('Input data must be 4D, '
                         'but provided data has shape {}'.format(input_data.shape))

    if input_data.shape[3] > 2:
        raise ValueError('Input data must have 1 or 2 channels, '
                         'but {} channels were provided'.format(input_data.shape[-1]))

    valid_channels = ['red', 'green', 'blue']
    channel_colors = [x.lower() for x in channel_colors]

    if not np.all(np.isin(channel_colors, valid_channels)):
        raise ValueError('Only red, green, or blue are valid channel colors')

    if len(channel_colors) != input_data.shape[-1]:
        raise ValueError('Must provide same number of channel_colors as channels in input_data')

    rgb_data = np.zeros(input_data.shape[:3] + (3,), dtype='float32')

    # rescale channels to aid plotting
    for img in range(input_data.shape[0]):
        for channel in range(input_data.shape[-1]):
            current_img = input_data[img, :, :, channel]
            non_zero_vals = current_img[np.nonzero(current_img)]

            # if there are non-zero pixels in current channel, we rescale
            if len(non_zero_vals) > 0:

                percentiles = np.percentile(non_zero_vals, [5, 95])
                rescaled_intensity = rescale_intensity(current_img,
                                                       in_range=(percentiles[0], percentiles[1]),
                                                       out_range='float32')

                # get rgb index of current channel
                color_idx = np.where(np.isin(valid_channels, channel_colors[channel]))
                rgb_data[img, :, :, color_idx] = rescaled_intensity

    # create a blank array for red channel
    return rgb_data


def make_outline_overlay(rgb_data, predictions):
    """Overlay a segmentation mask with image data for easy visualization

    Args:
        rgb_data: 3 channel array of images, output of ``create_rgb_data``
        predictions: segmentation predictions to be visualized

    Returns:
        numpy.array: overlay image of input data and predictions

    Raises:
        ValueError: If predictions are not 4D
        ValueError: If there is not matching RGB data for each prediction
    """
    if len(predictions.shape) != 4:
        raise ValueError('Predictions must be 4D, got {}'.format(predictions.shape))

    if predictions.shape[0] > rgb_data.shape[0]:
        raise ValueError('Must supply an rgb image for each prediction')

    boundaries = np.zeros_like(rgb_data)
    overlay_data = np.copy(rgb_data)

    for img in range(predictions.shape[0]):
        boundary = find_boundaries(predictions[img, ..., 0], connectivity=1, mode='inner')
        boundaries[img, boundary > 0, :] = 1

    overlay_data[boundaries > 0] = 1

    return overlay_data



# Copyright 2016-2021 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tf/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Custom Callbacks for DeepCell"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import timeit

import numpy as np

import tensorflow as tf
from tensorflow.keras import backend as K


class InferenceTimer(tf.keras.callbacks.Callback):
    """Callback to log inference speed per epoch."""

    def __init__(self, samples=100):
        super(InferenceTimer, self).__init__()
        self._samples = int(samples)
        self._batch_times = []
        self._samples_seen = []
        self._timer = None

    def on_predict_begin(self, epoch, logs=None):
        self._batch_times = []
        self._samples_seen = []

    def on_predict_batch_begin(self, batch, logs=None):
        self._timer = timeit.default_timer()

    def on_predict_batch_end(self, batch, logs=None):
        t = timeit.default_timer() - self._timer
        self._batch_times.append(t)
        outputs = logs.get('outputs', np.empty((1,)))
        if isinstance(self.model.output_shape, list):
            outputs = outputs[0]
        self._samples_seen.append(outputs.shape[0])

    def on_predict_end(self, logs=None):
        total_samples = np.sum(self._samples_seen)

        per_sample = [t / float(s) for t, s in
                      zip(self._batch_times, self._samples_seen)]

        avg = np.mean(per_sample)
        std = np.std(per_sample)

        print('Average inference speed per sample for %s total samples: '
              '%0.5fs ± %0.5fs.' % (total_samples, avg, std))

    def on_epoch_end(self, epoch, logs=None):
        shape = tuple([self._samples] + list(self.model.input_shape[1:]))
        test_batch = np.random.random(shape)
        self.model.predict(test_batch, callbacks=self)



# Copyright 2016-2021 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tf/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for NuclearSegmentationApplication"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.platform import test
import numpy as np

from deepcell.model_zoo import PanopticNet
from deepcell.applications import NuclearSegmentation


class TestNuclearSegmentation(test.TestCase):

    def test_nuclear_app(self):
        with self.cached_session():
            model = PanopticNet(
                'resnet50',
                input_shape=(128, 128, 1),
                norm_method='whole_image',
                num_semantic_heads=2,
                num_semantic_classes=[1, 1],
                location=True,
                include_top=True,
                lite=True,
                use_imagenet=False,
                interpolation='bilinear')
            app = NuclearSegmentation(model)

            # test output shape
            shape = app.model.output_shape
            self.assertIsInstance(shape, list)
            self.assertEqual(len(shape), 2)
            self.assertEqual(len(shape[0]), 4)
            self.assertEqual(len(shape[1]), 4)

            # test predict
            x = np.random.rand(1, 500, 500, 1)
            y = app.predict(x)
            self.assertEqual(x.shape, y.shape)



# Copyright 2016-2021 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tf/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""
Custom Image Data Generators used to generate augmented batches of training
data. These custom generators extend the keras.ImageDataGenerator, and allow
for training with label masks, bounding boxes, and more customized annotations.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import warnings

import numpy as np

from tensorflow.keras import backend as K
from tensorflow.keras.utils import to_categorical

from deepcell.utils import transform_utils


def _transform_masks(y, transform, data_format=None, **kwargs):
    """Based on the transform key, apply a transform function to the masks.

    Refer to :mod:`deepcell.utils.transform_utils` for more information about
    available transforms. Caution for unknown transform keys.

    Args:
        y (numpy.array): Labels of ``ndim`` 4 or 5
        transform (str): Name of the transform, one of
            ``{"deepcell", "disc", "watershed", None}``.
        data_format (str): A string, one of ``channels_last`` (default)
            or ``channels_first``. The ordering of the dimensions in the
            inputs. ``channels_last`` corresponds to inputs with shape
            ``(batch, height, width, channels)`` while ``channels_first``
            corresponds to inputs with shape
            ``(batch, channels, height, width)``.
        kwargs (dict): Optional transform keyword arguments.

    Returns:
        numpy.array: the output of the given transform function on ``y``.

    Raises:
        ValueError: Rank of ``y`` is not 4 or 5.
        ValueError: Channel dimension of ``y`` is not 1.
        ValueError: ``transform`` is invalid value.
    """
    valid_transforms = {
        'deepcell',  # deprecated for "pixelwise"
        'pixelwise',
        'disc',
        'watershed',  # deprecated for "outer-distance"
        'watershed-cont',  # deprecated for "outer-distance"
        'inner-distance', 'inner_distance',
        'outer-distance', 'outer_distance',
        'centroid',  # deprecated for "inner-distance"
        'fgbg'
    }
    if data_format is None:
        data_format = K.image_data_format()

    if y.ndim not in {4, 5}:
        raise ValueError('`labels` data must be of ndim 4 or 5.  Got', y.ndim)

    channel_axis = 1 if data_format == 'channels_first' else -1

    if y.shape[channel_axis] != 1:
        raise ValueError('Expected channel axis to be 1 dimension. Got',
                         y.shape[1 if data_format == 'channels_first' else -1])

    if isinstance(transform, str):
        transform = transform.lower()

    if transform not in valid_transforms and transform is not None:
        raise ValueError('`{}` is not a valid transform'.format(transform))

    if transform in {'pixelwise', 'deepcell'}:
        if transform == 'deepcell':
            warnings.warn('The `{}` transform is deprecated. Please use the '
                          '`pixelwise` transform instead.'.format(transform),
                          DeprecationWarning)
        dilation_radius = kwargs.pop('dilation_radius', None)
        separate_edge_classes = kwargs.pop('separate_edge_classes', False)

        edge_class_shape = 4 if separate_edge_classes else 3

        if data_format == 'channels_first':
            shape = tuple([y.shape[0]] + [edge_class_shape] + list(y.shape[2:]))
        else:
            shape = tuple(list(y.shape[0:-1]) + [edge_class_shape])

        # using uint8 since should only be 4 unique values.
        y_transform = np.zeros(shape, dtype=np.uint8)

        for batch in range(y_transform.shape[0]):
            if data_format == 'channels_first':
                mask = y[batch, 0, ...]
            else:
                mask = y[batch, ..., 0]

            y_transform[batch] = transform_utils.pixelwise_transform(
                mask, dilation_radius, data_format=data_format,
                separate_edge_classes=separate_edge_classes)

    elif transform in {'outer-distance', 'outer_distance',
                       'watershed', 'watershed-cont'}:
        if transform in {'watershed', 'watershed-cont'}:
            warnings.warn('The `{}` transform is deprecated. Please use the '
                          '`outer-distance` transform instead.'.format(transform),
                          DeprecationWarning)

        by_frame = kwargs.pop('by_frame', True)
        bins = kwargs.pop('distance_bins', None)

        distance_kwargs = {
            'bins': bins,
            'erosion_width': kwargs.pop('erosion_width', 0),
        }

        # If using 3d transform, pass in scale arg
        if y.ndim == 5 and not by_frame:
            distance_kwargs['sampling'] = kwargs.pop('sampling', [0.5, 0.217, 0.217])

        if data_format == 'channels_first':
            shape = tuple([y.shape[0]] + list(y.shape[2:]))
        else:
            shape = y.shape[0:-1]
        y_transform = np.zeros(shape, dtype=K.floatx())

        if y.ndim == 5:
            if by_frame:
                _distance_transform = transform_utils.outer_distance_transform_movie
            else:
                _distance_transform = transform_utils.outer_distance_transform_3d
        else:
            _distance_transform = transform_utils.outer_distance_transform_2d

        for batch in range(y_transform.shape[0]):
            if data_format == 'channels_first':
                mask = y[batch, 0, ...]
            else:
                mask = y[batch, ..., 0]
            y_transform[batch] = _distance_transform(mask, **distance_kwargs)

        y_transform = np.expand_dims(y_transform, axis=-1)

        if bins is not None:
            # convert to one hot notation
            # uint8's max value of255 seems like a generous limit for binning.
            y_transform = to_categorical(y_transform, num_classes=bins, dtype=np.uint8)
        if data_format == 'channels_first':
            y_transform = np.rollaxis(y_transform, y.ndim - 1, 1)

    elif transform in {'inner-distance', 'inner_distance', 'centroid'}:
        if transform == 'centroid':
            warnings.warn('The `{}` transform is deprecated. Please use the '
                          '`inner-distance` transform instead.'.format(transform),
                          DeprecationWarning)

        by_frame = kwargs.pop('by_frame', True)
        bins = kwargs.pop('distance_bins', None)

        distance_kwargs = {
            'bins': bins,
            'erosion_width': kwargs.pop('erosion_width', 0),
            'alpha': kwargs.pop('alpha', 0.1),
            'beta': kwargs.pop('beta', 1)
        }

        # If using 3d transform, pass in scale arg
        if y.ndim == 5 and not by_frame:
            distance_kwargs['sampling'] = kwargs.pop('sampling', [0.5, 0.217, 0.217])

        if data_format == 'channels_first':
            shape = tuple([y.shape[0]] + list(y.shape[2:]))
        else:
            shape = y.shape[0:-1]
        y_transform = np.zeros(shape, dtype=K.floatx())

        if y.ndim == 5:
            if by_frame:
                _distance_transform = transform_utils.inner_distance_transform_movie
            else:
                _distance_transform = transform_utils.inner_distance_transform_3d
        else:
            _distance_transform = transform_utils.inner_distance_transform_2d

        for batch in range(y_transform.shape[0]):
            if data_format == 'channels_first':
                mask = y[batch, 0, ...]
            else:
                mask = y[batch, ..., 0]
            y_transform[batch] = _distance_transform(mask, **distance_kwargs)

        y_transform = np.expand_dims(y_transform, axis=-1)

        if distance_kwargs['bins'] is not None:
            # convert to one hot notation
            # uint8's max value of255 seems like a generous limit for binning.
            y_transform = to_categorical(y_transform, num_classes=bins, dtype=np.uint8)
        if data_format == 'channels_first':
            y_transform = np.rollaxis(y_transform, y.ndim - 1, 1)

    elif transform == 'disc' or transform is None:
        dtype = K.floatx() if transform == 'disc' else np.int32
        y_transform = to_categorical(y.squeeze(channel_axis), dtype=dtype)
        if data_format == 'channels_first':
            y_transform = np.rollaxis(y_transform, y.ndim - 1, 1)

    elif transform == 'fgbg':
        y_transform = np.where(y > 1, 1, y)
        # convert to one hot notation
        if data_format == 'channels_first':
            y_transform = np.rollaxis(y_transform, 1, y.ndim)
        # using uint8 since should only be 2 unique values.
        y_transform = to_categorical(y_transform, dtype=np.uint8)
        if data_format == 'channels_first':
            y_transform = np.rollaxis(y_transform, y.ndim - 1, 1)

    return y_transform


# Globally-importable utils.
# pylint: disable=wrong-import-position
from deepcell.image_generators.fully_convolutional import ImageFullyConvDataGenerator
from deepcell.image_generators.fully_convolutional import ImageFullyConvIterator
from deepcell.image_generators.fully_convolutional import MovieDataGenerator
from deepcell.image_generators.fully_convolutional import MovieArrayIterator

from deepcell.image_generators.semantic import SemanticDataGenerator
from deepcell.image_generators.semantic import SemanticIterator
from deepcell.image_generators.semantic import SemanticMovieGenerator
from deepcell.image_generators.semantic import SemanticMovieIterator

from deepcell.image_generators.semantic import Semantic3DGenerator
from deepcell.image_generators.semantic import Semantic3DIterator

from deepcell.image_generators.sample import SampleDataGenerator
from deepcell.image_generators.sample import ImageSampleArrayIterator
from deepcell.image_generators.sample import SampleMovieDataGenerator
from deepcell.image_generators.sample import SampleMovieArrayIterator

from deepcell.image_generators.scale import ScaleIterator
from deepcell.image_generators.scale import ScaleDataGenerator

from deepcell.image_generators.tracking import SiameseDataGenerator
from deepcell.image_generators.tracking import SiameseIterator

from deepcell.image_generators.cropping import CroppingDataGenerator
from deepcell.image_generators.cropping import CroppingIterator
# pylint: enable=wrong-import-position

del absolute_import
del division
del print_function


__all__ = [
    'ImageFullyConvDataGenerator',
    'ImageFullyConvIterator',
    'MovieDataGenerator',
    'MovieArrayIterator',
    'SampleDataGenerator',
    'ImageSampleArrayIterator',
    'SampleMovieDataGenerator',
    'SampleMovieArrayIterator',
    'ScaleIterator',
    'ScaleDataGenerator',
    'SiameseDataGenerator',
    'SiameseIterator',
    'CroppingDataGenerator',
    'CroppingIterator',
    'SemanticIterator',
    'SemanticDataGenerator',
    'SemanticMovieIterator',
    'SemanticMovieGenerator',
    'Semantic3DIterator',
    'Semantic3DGenerator',
]



# Copyright 2016-2021 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tf/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for data_utils"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import json
import tarfile
import tempfile

import numpy as np
from tensorflow.keras import backend as K
from tensorflow.python.platform import test

from deepcell.utils import data_utils


class TestDataUtils(test.TestCase):

    def _write_test_trks(self, path):
        img_w, img_h = 30, 30
        X = np.random.random((10, img_w, img_h, 1))
        y = np.random.randint(4, size=(10, img_w, img_h, 1))
        lineages = [{
            1: {'daughters': [2, 3]},
            2: {'daughters': []},
            3: {'daughters': []}
        }] * X.shape[0]

        with tarfile.open(path, 'w') as trks:
            with tempfile.NamedTemporaryFile('w') as lineages_file:
                json.dump(lineages, lineages_file, indent=4)
                lineages_file.flush()
                trks.add(lineages_file.name, 'lineages.json')

            with tempfile.NamedTemporaryFile() as raw_file:
                np.save(raw_file, X)
                raw_file.flush()
                trks.add(raw_file.name, 'raw.npy')

            with tempfile.NamedTemporaryFile() as tracked_file:
                np.save(tracked_file, y)
                tracked_file.flush()
                trks.add(tracked_file.name, 'tracked.npy')

    def test_get_data(self):
        test_size = .1
        img_w, img_h = 30, 30
        X = np.random.random((10, img_w, img_h, 1))
        y = np.random.randint(3, size=(10, img_w, img_h, 1))

        temp_dir = self.get_temp_dir()
        good_file = os.path.join(temp_dir, 'good.npz')
        np.savez(good_file, X=X, y=y)

        train_dict, test_dict = data_utils.get_data(
            good_file, test_size=test_size)

        X_test, X_train = test_dict['X'], train_dict['X']

        self.assertIsInstance(train_dict, dict)
        self.assertIsInstance(test_dict, dict)
        self.assertAlmostEqual(X_test.size / (X_test.size + X_train.size), test_size)

        # test bad filepath
        bad_file = os.path.join(temp_dir, 'bad.npz')
        np.savez(bad_file, X_bad=X, y_bad=y)
        with self.assertRaises(KeyError):
            _, _ = data_utils.get_data(bad_file)

        # test siamese_daughters mode
        good_file = os.path.join(temp_dir, 'siamese.trks')
        self._write_test_trks(good_file)

        train_dict, test_dict = data_utils.get_data(
            good_file, mode='siamese_daughters', test_size=test_size)

        X_test, X_train = test_dict['X'], train_dict['X']

        d_test, d_train = test_dict['daughters'], train_dict['daughters']

        self.assertIsInstance(train_dict, dict)
        self.assertIsInstance(test_dict, dict)
        self.assertIsInstance(d_test, list)
        self.assertIsInstance(d_train, list)
        self.assertEqual(len(d_train), X_train.shape[0])
        self.assertEqual(len(d_test), X_test.shape[0])
        self.assertAlmostEqual(X_test.size / (X_test.size + X_train.size), test_size)

    def test_load_trks(self):
        temp_dir = self.get_temp_dir()
        good_file = os.path.join(temp_dir, 'siamese.trks')
        self._write_test_trks(good_file)

        trks = data_utils.load_trks(good_file)
        X = trks.get('X')
        y = trks.get('y')
        lineages = trks.get('lineages')
        self.assertIsInstance(trks, dict)
        self.assertIsInstance(trks.get('X'), np.ndarray)
        self.assertIsInstance(trks.get('y'), np.ndarray)
        self.assertIsInstance(trks.get('lineages'), list)
        self.assertEqual(len(lineages), X.shape[0])
        for i in range(y.shape[0]):
            expected_keys = sorted([u for u in np.unique(y[i]) if u != 0])
            print(expected_keys)
            self.assertAllEqual(sorted(lineages[i].keys()), expected_keys)
            for k in lineages[i]:
                self.assertIsInstance(lineages[i][k]['daughters'], list)

    def test_get_max_sample_num_list(self):
        K.set_image_data_format('channels_last')
        edge_feature = [1, 0, 0]  # first channel index is cell edge
        win_x, win_y = 10, 10
        y = np.zeros((2, 30, 30, 3))
        y[:, 0, 0, 0] = 1  # set value outside valid window range
        y[:, win_x + 1, win_y + 1, 0] = 1  # set value inside valid window range
        # test non-sample mode
        max_nums = data_utils.get_max_sample_num_list(y, edge_feature,
                                                      output_mode='conv',
                                                      padding='same',
                                                      window_size_x=win_x,
                                                      window_size_y=win_y)
        self.assertEqual(max_nums, [np.Inf, np.Inf])

        # test sample mode, no padding
        max_nums = data_utils.get_max_sample_num_list(y, edge_feature,
                                                      output_mode='sample',
                                                      padding='same',
                                                      window_size_x=win_x,
                                                      window_size_y=win_y)
        self.assertEqual(max_nums, [2, 2])

        # test sample mode, valid padding
        max_nums = data_utils.get_max_sample_num_list(y, edge_feature,
                                                      output_mode='sample',
                                                      padding='valid',
                                                      window_size_x=win_x,
                                                      window_size_y=win_y)
        self.assertEqual(max_nums, [1, 1])

        # channels_first
        K.set_image_data_format('channels_first')
        edge_feature = [1, 0, 0]  # first channel index is cell edge
        win_x, win_y = 10, 10
        y = np.zeros((2, 3, 30, 30))
        y[:, 0, 0, 0] = 1  # set value outside valid window range
        y[:, 0, win_x + 1, win_y + 1] = 1  # set value inside valid window range

        # test sample mode, no padding
        max_nums = data_utils.get_max_sample_num_list(y, edge_feature,
                                                      output_mode='sample',
                                                      padding='same',
                                                      window_size_x=win_x,
                                                      window_size_y=win_y)
        self.assertEqual(max_nums, [2, 2])

        # test sample mode, valid padding
        max_nums = data_utils.get_max_sample_num_list(y, edge_feature,
                                                      output_mode='sample',
                                                      padding='valid',
                                                      window_size_x=win_x,
                                                      window_size_y=win_y)
        self.assertEqual(max_nums, [1, 1])

    def test_sample_label_matrix(self):
        win_x, win_y = 10, 10
        y = np.zeros((2, 30, 30, 3))
        y[:, 0, 0, 0] = 1  # set value outside valid window range
        y[:, win_x + 1, win_y + 1, 0] = 1  # set value inside valid window range
        r, c, b, l = data_utils.sample_label_matrix(
            y, window_size=(win_x, win_y),
            padding='valid', data_format='channels_last')
        self.assertListEqual(list(map(len, [r, c, b, l])), [len(r)] * 4)
        self.assertEqual(np.unique(b).size, 2)
        self.assertEqual([np.unique(r).size, np.unique(c).size], [1, 1])
        self.assertEqual(np.unique(l).size, 1)

        r, c, b, l = data_utils.sample_label_matrix(
            y, window_size=(win_x, win_y),
            max_training_examples=None,
            padding='same', data_format='channels_last')
        self.assertListEqual(list(map(len, [r, c, b, l])), [len(r)] * 4)
        self.assertEqual([np.unique(r).size, np.unique(c).size], [2, 2])
        self.assertEqual(np.unique(l).size, 1)

        # test channels_first
        win_x, win_y = 10, 10
        y = np.zeros((2, 3, 30, 30))
        y[:, 0, 0, 0] = 1  # set value outside valid window range
        y[:, 0, win_x + 1, win_y + 1] = 1  # set value inside valid window range
        r, c, b, l = data_utils.sample_label_matrix(
            y, window_size=(win_x, win_y),
            padding='valid', data_format='channels_first')
        self.assertListEqual(list(map(len, [r, c, b, l])), [len(r)] * 4)
        self.assertEqual(np.unique(b).size, 2)
        self.assertEqual([np.unique(r).size, np.unique(c).size], [1, 1])
        self.assertEqual(np.unique(l).size, 1)

        r, c, b, l = data_utils.sample_label_matrix(
            y, window_size=(win_x, win_y),
            max_training_examples=None,
            padding='same', data_format='channels_first')
        self.assertListEqual(list(map(len, [r, c, b, l])), [len(r)] * 4)
        self.assertEqual(np.unique(b).size, 2)
        self.assertEqual([np.unique(r).size, np.unique(c).size], [2, 2])
        self.assertEqual(np.unique(l).size, 1)

    def test_sample_label_movie(self):
        win_x, win_y, win_z = 10, 10, 1
        y = np.zeros((2, 5, 30, 30, 3))
        # set value outside valid window range
        y[:, 0, 0, 0, 0] = 1
        # set value inside valid window range
        y[:, win_z + 1, win_x + 1, win_y + 1, 0] = 1
        f, r, c, b, l = data_utils.sample_label_movie(
            y, window_size=(win_x, win_y, win_z),
            padding='valid', data_format='channels_last')
        self.assertListEqual(list(map(len, [r, c, b, l, f])), [len(r)] * 5)
        self.assertEqual(np.unique(b).size, 2)
        self.assertEqual([np.unique(r).size,
                          np.unique(c).size,
                          np.unique(f).size],
                         [1, 1, 1])
        self.assertEqual(np.unique(l).size, 1)

        f, r, c, b, l = data_utils.sample_label_movie(
            y, window_size=(win_x, win_y, win_z),
            max_training_examples=None,
            padding='same', data_format='channels_last')
        self.assertListEqual(list(map(len, [r, c, b, l, f])), [len(r)] * 5)
        self.assertEqual([np.unique(r).size,
                          np.unique(c).size,
                          np.unique(f).size],
                         [2, 2, 2])
        self.assertEqual(np.unique(l).size, 1)

        # test channels_first
        win_x, win_y, win_z = 10, 10, 1
        y = np.zeros((2, 3, 5, 30, 30))
        # set value outside valid window range
        y[:, 0, 0, 0, 0] = 1
        # set value inside valid window range
        y[:, 0, win_z + 1, win_x + 1, win_y + 1] = 1
        f, r, c, b, l = data_utils.sample_label_movie(
            y, window_size=(win_x, win_y, win_z),
            padding='valid', data_format='channels_first')
        self.assertListEqual(list(map(len, [r, c, b, l, f])), [len(r)] * 5)
        self.assertEqual(np.unique(b).size, 2)
        self.assertEqual([np.unique(r).size, np.unique(c).size], [1, 1])
        self.assertEqual(np.unique(l).size, 1)

        f, r, c, b, l = data_utils.sample_label_movie(
            y, window_size=(win_x, win_y, win_z),
            max_training_examples=None,
            padding='same', data_format='channels_first')
        self.assertListEqual(list(map(len, [r, c, b, l, f])), [len(r)] * 5)
        self.assertEqual(np.unique(b).size, 2)
        self.assertEqual([np.unique(r).size, np.unique(c).size], [2, 2])
        self.assertEqual(np.unique(l).size, 1)

    def test_trim_padding(self):
        # test 2d image
        K.set_image_data_format('channels_last')
        img_size = 512
        win_x, win_y = 30, 30
        x_trim = img_size - 2 * win_x
        y_trim = img_size - 2 * win_y
        K.set_image_data_format('channels_last')
        arr = np.zeros((1, img_size, img_size, 1))
        arr_trim = data_utils.trim_padding(arr, win_x, win_y)
        self.assertEqual(arr_trim.shape, (1, x_trim, y_trim, 1))
        # test channels_first
        K.set_image_data_format('channels_first')
        arr = np.zeros((1, 1, img_size, img_size))
        arr_trim = data_utils.trim_padding(arr, win_x, win_y)
        self.assertEqual(arr_trim.shape, (1, 1, x_trim, y_trim))

        # test 3d image stack
        img_size = 256
        frames = 30
        win_x, win_y = 20, 30
        win_z = 2
        x_trim = img_size - 2 * win_x
        y_trim = img_size - 2 * win_y
        z_trim = frames - 2 * win_z
        K.set_image_data_format('channels_last')
        arr = np.zeros((1, frames, img_size, img_size, 1))
        # trim win_z
        arr_trim = data_utils.trim_padding(arr, win_x, win_y, win_z)
        self.assertEqual(arr_trim.shape, (1, z_trim, x_trim, y_trim, 1))
        # don't trim win_z
        arr_trim = data_utils.trim_padding(arr, win_x, win_y)
        self.assertEqual(arr_trim.shape, (1, frames, x_trim, y_trim, 1))
        # test channels_first
        K.set_image_data_format('channels_first')
        arr = np.zeros((1, 1, 30, img_size, img_size))
        # trim win_z
        arr_trim = data_utils.trim_padding(arr, win_x, win_y, win_z)
        self.assertEqual(arr_trim.shape, (1, 1, z_trim, x_trim, y_trim))
        # don't trim win_z
        arr_trim = data_utils.trim_padding(arr, win_x, win_y)
        self.assertEqual(arr_trim.shape, (1, 1, frames, x_trim, y_trim))

        # test bad input
        with self.assertRaises(ValueError):
            small_arr = np.zeros((img_size, img_size, 1))
            data_utils.trim_padding(small_arr, 10, 10)
        with self.assertRaises(ValueError):
            big_arr = np.zeros((1, 1, 30, img_size, img_size, 1))
            data_utils.trim_padding(big_arr, 10, 10)

    def test_relabel_movie(self):
        y = np.array([[0, 3, 5], [4, 99, 123]])
        relabeled = data_utils.relabel_movie(y)
        self.assertAllEqual(relabeled, np.array([[0, 1, 3], [2, 4, 5]]))

    def test_reshape_movie(self):
        K.set_image_data_format('channels_last')
        batches = np.random.randint(1, 5)
        L = 16
        frames = 3
        channels = 3
        X = np.zeros((batches, frames, L, L, channels))
        y = np.random.randint(low=0, high=1000, size=(batches, frames, L, L, 1))
        new_size = 4
        # guarantee there is a 0
        y[:, 0, 0, 0, 0] = 0

        # test resize to smaller image, divisible
        new_X, new_y = data_utils.reshape_movie(X, y, new_size)
        new_batch = np.ceil(L / new_size) ** 2 * batches
        self.assertEqual(new_X.shape,
                         (new_batch, frames, new_size, new_size, channels))
        self.assertEqual(new_y.shape,
                         (new_batch, frames, new_size, new_size, 1))
        for b in range(new_X.shape[0]):
            self.assertEqual(list(np.unique(new_y[b])),
                             list(range(new_y[b].max() + 1)))

        # test reshape with non-divisible values.
        new_size = 5
        new_batch = np.ceil(L / new_size) ** 2 * batches
        new_X, new_y = data_utils.reshape_movie(X, y, new_size)
        self.assertEqual(new_X.shape,
                         (new_batch, frames, new_size, new_size, channels))
        self.assertEqual(new_y.shape,
                         (new_batch, frames, new_size, new_size, 1))
        for b in range(new_X.shape[0]):
            self.assertEqual(list(np.unique(new_y[b])),
                             list(range(new_y[b].max() + 1)))

        # test reshape to bigger size
        with self.assertRaises(ValueError):
            new_X, new_y = data_utils.reshape_movie(X, y, L * 2)

        # test wrong dimensions
        bigger = np.zeros((1, frames, L, L, frames, 1))
        smaller = np.zeros((1, L, L, frames))
        with self.assertRaises(ValueError):
            new_X, new_y = data_utils.reshape_movie(smaller, y, new_size)
        with self.assertRaises(ValueError):
            new_X, new_y = data_utils.reshape_movie(bigger, y, new_size)
        with self.assertRaises(ValueError):
            new_X, new_y = data_utils.reshape_movie(X, smaller, new_size)
        with self.assertRaises(ValueError):
            new_X, new_y = data_utils.reshape_movie(X, bigger, new_size)

        # channels_first
        K.set_image_data_format('channels_first')
        X = np.zeros((1, channels, frames, L, L))
        y = np.zeros((1, 1, frames, L, L))
        new_size = 4

        # test resize to smaller image, divisible
        new_X, new_y = data_utils.reshape_movie(X, y, new_size)
        new_batch = np.ceil(L / new_size) ** 2
        self.assertEqual(new_X.shape,
                         (new_batch, channels, frames, new_size, new_size))
        self.assertEqual(new_y.shape,
                         (new_batch, 1, frames, new_size, new_size))
        for b in range(new_X.shape[0]):
            self.assertEqual(list(np.unique(new_y[b])),
                             list(range(new_y[b].max() + 1)))

        # test reshape with non-divisible values.
        new_size = 5
        new_batch = np.ceil(L / new_size) ** 2
        new_X, new_y = data_utils.reshape_movie(X, y, new_size)
        self.assertEqual(new_X.shape,
                         (new_batch, channels, frames, new_size, new_size))
        self.assertEqual(new_y.shape,
                         (new_batch, 1, frames, new_size, new_size))
        for b in range(new_X.shape[0]):
            self.assertEqual(list(np.unique(new_y[b])),
                             list(range(new_y[b].max() + 1)))

    def test_reshape_matrix(self):
        K.set_image_data_format('channels_last')
        batches = np.random.randint(1, 5)
        L = 8
        channels = 3
        X = np.zeros((batches, L, L, channels))
        y = np.random.randint(low=0, high=1000, size=(batches, L, L, 1))
        new_size = 4
        # guarantee there is a 0
        y[:, 0, 0, 0] = 0

        # test resize to smaller image, divisible
        new_X, new_y = data_utils.reshape_matrix(X, y, new_size)
        new_batch = np.ceil(L / new_size) ** 2 * batches
        self.assertEqual(new_X.shape, (new_batch, new_size, new_size, channels))
        self.assertEqual(new_y.shape, (new_batch, new_size, new_size, 1))
        for b in range(new_y.shape[0]):
            self.assertEqual(list(np.unique(new_y[b])),
                             list(range(new_y[b].max() + 1)))

        # test reshape with non-divisible values.
        new_size = 5
        new_batch = np.ceil(L / new_size) ** 2 * batches
        new_X, new_y = data_utils.reshape_matrix(X, y, new_size)
        self.assertEqual(new_X.shape, (new_batch, new_size, new_size, channels))
        self.assertEqual(new_y.shape, (new_batch, new_size, new_size, 1))
        for b in range(new_y.shape[0]):
            self.assertEqual(list(np.unique(new_y[b])),
                             list(range(new_y[b].max() + 1)))

        # test reshape to bigger size
        with self.assertRaises(ValueError):
            new_X, new_y = data_utils.reshape_matrix(X, y, L * 2)

        # test wrong dimensions
        bigger = np.zeros((1, L, L, channels, 1))
        smaller = np.zeros((1, L, L))
        with self.assertRaises(ValueError):
            new_X, new_y = data_utils.reshape_matrix(smaller, y, new_size)
        with self.assertRaises(ValueError):
            new_X, new_y = data_utils.reshape_matrix(bigger, y, new_size)
        with self.assertRaises(ValueError):
            new_X, new_y = data_utils.reshape_matrix(X, smaller, new_size)
        with self.assertRaises(ValueError):
            new_X, new_y = data_utils.reshape_matrix(X, bigger, new_size)

        # channels_first
        K.set_image_data_format('channels_first')
        X = np.zeros((1, channels, L, L))
        y = np.zeros((1, 1, L, L))
        new_size = 4

        # test resize to smaller image, divisible
        new_X, new_y = data_utils.reshape_matrix(X, y, new_size)
        new_batch = np.ceil(L / new_size) ** 2
        self.assertEqual(new_X.shape, (new_batch, channels, new_size, new_size))
        self.assertEqual(new_y.shape, (new_batch, 1, new_size, new_size))
        for b in range(new_X.shape[0]):
            self.assertEqual(list(np.unique(new_y[b])),
                             list(range(new_y[b].max() + 1)))

        # test reshape with non-divisible values.
        new_size = 5
        new_batch = np.ceil(L / new_size) ** 2
        new_X, new_y = data_utils.reshape_matrix(X, y, new_size)
        self.assertEqual(new_X.shape, (new_batch, channels, new_size, new_size))
        self.assertEqual(new_y.shape, (new_batch, 1, new_size, new_size))
        for b in range(new_X.shape[0]):
            self.assertEqual(list(np.unique(new_y[b])),
                             list(range(new_y[b].max() + 1)))

        # test reshape with non-square image, square new size
        K.set_image_data_format('channels_last')
        batches = np.random.randint(1, 5)
        Lx, Ly = 30, 40
        channels = 3
        X = np.zeros((batches, Lx, Ly, channels))
        y = np.random.randint(low=0, high=1000, size=(batches, Lx, Ly, 1))
        new_size = 4
        new_batch = batches * np.ceil(Lx / new_size) * np.ceil(Ly / new_size)
        new_X, new_y = data_utils.reshape_matrix(X, y, new_size)
        self.assertEqual(new_X.shape, (new_batch, new_size, new_size, channels))
        self.assertEqual(new_y.shape, (new_batch, new_size, new_size, 1))

        # test reshape with non-square image, non-square new size
        K.set_image_data_format('channels_last')
        batches = np.random.randint(1, 5)
        Lx, Ly = 30, 40
        channels = 3
        X = np.zeros((batches, Lx, Ly, channels))
        y = np.random.randint(low=0, high=1000, size=(batches, Lx, Ly, 1))
        new_size = [15, 10]
        new_batch = batches * np.ceil(Lx / new_size[0]) * np.ceil(Ly / new_size[1])
        new_X, new_y = data_utils.reshape_matrix(X, y, new_size)
        self.assertEqual(new_X.shape, (new_batch, new_size[0], new_size[1], channels))
        self.assertEqual(new_y.shape, (new_batch, new_size[0], new_size[1], 1))


if __name__ == '__main__':
    test.main()



# Copyright 2016-2021 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tf/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""A model that can detect whether 2 cells are same, different, or related."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

import tensorflow as tf

import deepcell_tracking
from deepcell_toolbox.processing import normalize

from deepcell.applications import Application


MODEL_PATH = ('https://deepcell-data.s3-us-west-1.amazonaws.com/'
              'saved-models/TrackingModel-4.tar.gz')

ENCODER_PATH = ('https://deepcell-data.s3-us-west-1.amazonaws.com/'
                'saved-models/TrackingModelNE-2.tar.gz')


class CellTracking(Application):
    """Loads a :mod:`deepcell.model_zoo.tracking.GNNTrackingModel` model for
    object tracking with pretrained weights using a simple
    ``predict`` interface.

    Args:
        use_pretrained_weights (bool): Whether to load pretrained weights.
        model_image_shape (tuple): Shape of input data expected by model.
        neighborhood_scale_size (int): Size of the area surrounding each cell.
        birth (float): Cost of new cell in linear assignment matrix.
        death (float): Cost of cell death in linear assignment matrix.
        division (float): Cost of cell division in linear assignment matrix.
    """

    #: Metadata for the dataset used to train the model
    dataset_metadata = {
        'name': 'tracked_nuclear_train_large',
        'other': 'Pooled tracked nuclear data from HEK293, HeLa-S3, NIH-3T3, and RAW264.7 cells.'
    }

    #: Metadata for the model and training process
    model_metadata = {
        'batch_size': 4,
        'track_length': 8,
        'lr': 1e-3,
        'clipnorm': 0.001,
        'n_epochs': 30,
        'training_steps_per_epoch': 512,
        'validation_steps': 100,
        'min_lr': 1e-7,
        'training_seed': None,
        'crop_dim': 32
    }

    def __init__(self,
                 model=None,
                 neighborhood_encoder=None,
                 distance_threshold=64,
                 birth=0.99,
                 death=0.99,
                 division=0.9,
                 track_length=8,
                 embedding_axis=0):
        self.neighborhood_encoder = neighborhood_encoder
        self.distance_threshold = distance_threshold
        self.birth = birth
        self.death = death
        self.division = division
        self.track_length = track_length
        self.embedding_axis = embedding_axis

        if self.neighborhood_encoder is None:
            archive_path = tf.keras.utils.get_file(
                'TrackingModelNE.tgz', ENCODER_PATH,
                file_hash='c02a077965a3cd8ba4ca55bfc976caf1',
                extract=True, cache_subdir='models')
            model_path = os.path.splitext(archive_path)[0]
            self.neighborhood_encoder = tf.keras.models.load_model(model_path)

        if model is None:
            archive_path = tf.keras.utils.get_file(
                'TrackingModelInf.tgz', MODEL_PATH,
                file_hash='fd59fa1a36f9df6138c34d533bc73604',
                extract=True, cache_subdir='models')
            model_path = os.path.splitext(archive_path)[0]
            model = tf.keras.models.load_model(model_path)

        super(CellTracking, self).__init__(
            model,
            model_mpp=0.65,
            preprocessing_fn=None,
            postprocessing_fn=None,
            dataset_metadata=self.dataset_metadata,
            model_metadata=self.model_metadata)

    def predict(self, image, labels, **kwargs):
        """Using both raw image data and segmentation masks,
        track objects across all frames.

        Args:
            image (numpy.array): Raw image data.
            labels (numpy.array): Labels for ``image``, integer masks.

        Returns:
            dict: Tracked labels and lineage information.
        """
        image_norm = normalize(image)

        cell_tracker = deepcell_tracking.CellTracker(
            image_norm, labels, self.model,
            neighborhood_encoder=self.neighborhood_encoder,
            distance_threshold=self.distance_threshold,
            track_length=self.track_length,
            embedding_axis=self.embedding_axis,
            birth=self.birth, death=self.death,
            division=self.division)

        cell_tracker.track_cells()

        return cell_tracker._track_review_dict()

    def track(self, image, labels, **kwargs):
        """Wrapper around predict() for convenience."""
        return self.predict(image, labels, **kwargs)



# Copyright 2016-2021 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tf/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Detect the scale of input data for rescaling for other models"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

import numpy as np
import tensorflow as tf

from deepcell.applications import Application
from deepcell.layers import ImageNormalization2D
from deepcell.layers import TensorProduct
from deepcell.utils.backbone_utils import get_backbone


MODEL_PATH = ('https://deepcell-data.s3-us-west-1.amazonaws.com/'
              'saved-models/ScaleDetection-1.tar.gz')


def ScaleDetectionModel(input_shape=(None, None, 1),
                        inputs=None,
                        backbone='mobilenetv2'):
    """Create a ``ScaleDetectionModel`` for detecting scales of input data.

    This enables data to be scaled appropriately for other segmentation models
    which may not be resolution tolerant.

    Based on a standard backbone with an intiial ``ImageNormalization2D`` and
    final ``AveragePooling2D`` and ``TensorProduct`` layers.

    Args:
        input_shape (tuple): a 3-length tuple of the input data shape.
        inputs (tensorflow.keras.Layer): Optional input layer of the model.
            If not provided, creates a ``Layer`` based on ``input_shape``.
        backbone (str): name of the backbone to use for the model.
    """
    required_channels = 3  # required for most backbones

    if inputs is None:
        inputs = tf.keras.layers.Input(shape=input_shape)

    if tf.keras.backend.image_data_format() == 'channels_first':
        channel_axis = 0
    else:
        channel_axis = -1

    norm = ImageNormalization2D(norm_method='whole_image')(inputs)
    fixed_inputs = TensorProduct(required_channels)(norm)

    # force the input shape
    fixed_input_shape = list(input_shape)
    fixed_input_shape[channel_axis] = required_channels
    fixed_input_shape = tuple(fixed_input_shape)

    backbone_model = get_backbone(
        backbone,
        fixed_inputs,
        use_imagenet=False,
        return_dict=False,
        include_top=False,
        weights=None,
        input_shape=fixed_input_shape,
        pooling=None)

    x = tf.keras.layers.AveragePooling2D(4)(backbone_model.outputs[0])
    x = tf.keras.layers.Flatten()(x)
    x = TensorProduct(256, activation='relu')(x)
    outputs = TensorProduct(1, dtype=tf.keras.backend.floatx())(x)

    model = tf.keras.Model(inputs=backbone_model.inputs, outputs=outputs)

    return model


class ScaleDetection(Application):
    """Loads a :mod:`~ScaleDetectionModel` for detecting relative scales
    of images.

    Args:
        model (tf.keras.Model): The model to load. If ``None``,
            a pre-trained model will be downloaded.
    """

    #: Metadata for the dataset used to train the model
    dataset_metadata = {
        'name': 'general_nuclear_and_cyto_large',
        'other': 'Collection of all available nuclear and cytplasm stains.'
    }

    #: Metadata for the model and training process
    model_metadata = {
        'batch_size': 64,
        'lr': 1e-3,
        'lr_decay': 0.9,
        'training_seed': 0,
        'n_epochs': 200,
        'training_steps_per_epoch': 8400,
        'validation_steps_per_epoch': 2102,
        'error_rate': .01
    }

    def __init__(self, model=None):

        if model is None:
            archive_path = tf.keras.utils.get_file(
                'ScaleDetection.tgz', MODEL_PATH,
                file_hash='1ca4ee3a90fd4445bb5484e10ac0081c',
                extract=True, cache_subdir='models'
            )
            model_path = os.path.splitext(archive_path)[0]
            model = tf.keras.models.load_model(model_path)

        super(ScaleDetection, self).__init__(
            model,
            model_image_shape=model.input_shape[1:],
            model_mpp=0.65,
            preprocessing_fn=None,
            postprocessing_fn=None,
            dataset_metadata=self.dataset_metadata,
            model_metadata=self.model_metadata)

    def predict(self,
                image,
                batch_size=4,
                image_mpp=None):
        """Generates a labeled image of the input running prediction with
        appropriate pre and post processing functions.

        Input images are required to have 4 dimensions
        ``[batch, x, y, channel]``.
        Additional empty dimensions can be added using ``np.expand_dims``.

        Args:
            image (numpy.array): Input image with shape
                ``[batch, x, y, channel]``.
            batch_size (int): Number of images to predict on per batch.
            image_mpp (float): Microns per pixel for ``image``.

        Raises:
            ValueError: Input data must match required rank of the application,
                calculated as one dimension more (batch dimension) than expected
                by the model.

            ValueError: Input data must match required number of channels.

        Returns:
            numpy.array: Labeled image
            numpy.array: Model output
        """

        # Check input size of image
        if len(image.shape) != self.required_rank:
            raise ValueError('Input data must have {} dimensions. '
                             'Input data only has {} dimensions'.format(
                                 self.required_rank, len(image.shape)))

        if image.shape[-1] != self.required_channels:
            raise ValueError('Input data must have {} channels. '
                             'Input data only has {} channels'.format(
                                 self.required_channels, image.shape[-1]))

        # Resize image, returns unmodified if appropriate
        resized_image = self._resize_input(image, image_mpp)

        # Tile images, raises error if the image is not 4d
        tiles, _ = self._tile_input(resized_image)

        # Run images through model
        scales = self.model.predict(tiles, batch_size=batch_size)

        detected_scale = np.mean(scales)

        error_rate = self.model_metadata['error_rate']
        if abs(detected_scale - 1) < error_rate:
            detected_scale = 1

        return detected_scale



# Copyright 2016-2021 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tf/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Functions for running convolutional neural networks"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import numpy as np
from tensorflow.keras import backend as K
from tensorflow.keras.models import Model

from deepcell.utils.data_utils import trim_padding


def get_cropped_input_shape(images,
                            num_crops=4,
                            receptive_field=61,
                            data_format=None):
    """Calculate the input_shape for models to process cropped sub-images.

    Args:
        images (numpy.array): numpy array of original data
        num_crops (int): number of slices for the x and y axis
            to create sub-images
        receptive_field (int): the receptive field of the neural network.
        data_format (str): "channels_first" or "channels_last"

    Returns:
        tuple: new ``input_shape`` for model to process sub-images.
    """
    if data_format is None:
        data_format = K.image_data_format()
    if data_format == 'channels_first':
        channel_axis = 1
        row_axis = len(images.shape) - 2
        col_axis = len(images.shape) - 1
    else:
        channel_axis = len(images.shape) - 1
        row_axis = len(images.shape) - 3
        col_axis = len(images.shape) - 2

    channel_dim = images.shape[channel_axis]

    # Split the frames into quarters, as the full image size is too large
    crop_x = images.shape[row_axis] // num_crops + (receptive_field - 1)
    crop_y = images.shape[col_axis] // num_crops + (receptive_field - 1)

    if images.ndim == 5:
        input_shape = (images.shape[row_axis - 1], crop_x, crop_y, channel_dim)
    else:
        input_shape = (crop_x, crop_y, channel_dim)

    # switch to channels_first if necessary
    if channel_axis == 1:
        input_shape = tuple([input_shape[-1]] + list(input_shape[:-1]))

    return input_shape


def get_padding_layers(model):
    """Get all names of padding layers in a model

    Args:
        model (tensorflow.keras.Model): Keras model

    Returns:
        list: list of names of padding layers inside model
    """
    padding_layers = []
    for layer in model.layers:
        if 'padding' in layer.name:
            padding_layers.append(layer.name)
        elif isinstance(layer, Model):
            padding_layers.extend(get_padding_layers(layer))
    return padding_layers


def process_whole_image(model, images, num_crops=4, receptive_field=61, padding=None):
    """Slice images into num_crops * num_crops pieces, and use the model to
    process each small image.

    Args:
        model (tensorflow.keras.Model): model that will process each small image
        images (numpy.array): numpy array that is too big for model.predict
        num_crops (int): number of slices for the x and y axis
            to create sub-images
        receptive_field (int): receptive field used by model,
            required to pad images
        padding (str): type of padding for input images,
            one of {'reflect', 'zero'}.

    Returns:
        numpy.array: model outputs for each sub-image

    Raises:
        ValueError: invalid padding value
        ValueError: model input shape is different than expected_input_shape
    """
    if K.image_data_format() == 'channels_first':
        channel_axis = 1
        row_axis = len(images.shape) - 2
        col_axis = len(images.shape) - 1
    else:
        channel_axis = len(images.shape) - 1
        row_axis = len(images.shape) - 3
        col_axis = len(images.shape) - 2

    if not padding:
        padding_layers = get_padding_layers(model)
        if padding_layers:
            padding = 'reflect' if 'reflect' in padding_layers[0] else 'zero'

    if str(padding).lower() not in {'reflect', 'zero'}:
        raise ValueError('Expected `padding_mode` to be either `zero` or '
                         '`reflect`.  Got ', padding)

    # Split the frames into quarters, as the full image size is too large
    crop_x = images.shape[row_axis] // num_crops
    crop_y = images.shape[col_axis] // num_crops

    # Set up receptive field window for padding
    win_x, win_y = (receptive_field - 1) // 2, (receptive_field - 1) // 2

    # instantiate matrix for model output
    model_output_shape = tuple(list(model.layers[-1].output_shape)[1:])
    if channel_axis == 1:
        output = np.zeros(tuple([images.shape[0], model_output_shape[0]] +
                                list(images.shape[2:])))
    else:
        output = np.zeros(tuple(list(images.shape[0:-1]) +
                                [model_output_shape[-1]]))

    expected_input_shape = get_cropped_input_shape(
        images, num_crops, receptive_field)

    if expected_input_shape != model.input_shape[1:]:
        raise ValueError('Expected model.input_shape to be {}. Got {}.  Use '
                         '`get_cropped_input_shape()` to recreate your model '
                         ' with the proper input_shape'.format(
                             expected_input_shape, model.input_shape[1:]))

    # pad the images only in the x and y axes
    pad_width = []
    for i in range(len(images.shape)):
        if i == row_axis:
            pad_width.append((win_x, win_x))
        elif i == col_axis:
            pad_width.append((win_y, win_y))
        else:
            pad_width.append((0, 0))

    if str(padding).lower() == 'reflect':
        padded_images = np.pad(images, pad_width, mode='reflect')
    else:
        padded_images = np.pad(images, pad_width, mode='constant', constant_values=0)

    for i in range(num_crops):
        for j in range(num_crops):
            e, f = i * crop_x, (i + 1) * crop_x + 2 * win_x
            g, h = j * crop_y, (j + 1) * crop_y + 2 * win_y

            if images.ndim == 5:
                if channel_axis == 1:
                    predicted = model.predict(padded_images[:, :, :, e:f, g:h])
                else:
                    predicted = model.predict(padded_images[:, :, e:f, g:h, :])
            else:
                if channel_axis == 1:
                    predicted = model.predict(padded_images[:, :, e:f, g:h])
                else:
                    predicted = model.predict(padded_images[:, e:f, g:h, :])

            # if using skip_connections, get the final model output
            if isinstance(predicted, list):
                predicted = predicted[-1]

            # if the model uses padding, trim the output images to proper shape
            # if model does not use padding, images should already be correct
            if padding:
                predicted = trim_padding(predicted, win_x, win_y)

            a, b = i * crop_x, (i + 1) * crop_x
            c, d = j * crop_y, (j + 1) * crop_y

            if images.ndim == 5:
                if channel_axis == 1:
                    output[:, :, :, a:b, c:d] = predicted
                else:
                    output[:, :, a:b, c:d, :] = predicted
            else:
                if channel_axis == 1:
                    output[:, :, a:b, c:d] = predicted
                else:
                    output[:, a:b, c:d, :] = predicted

    return output



# Copyright 2016-2021 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tf/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Semantic segmentation data generators with cropping."""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import os
import warnings

import numpy as np

from tensorflow.python.keras.preprocessing.image import array_to_img

from deepcell.image_generators import SemanticDataGenerator, SemanticIterator

try:
    import scipy
    # scipy.linalg cannot be accessed until explicitly imported
    from scipy import linalg
    # scipy.ndimage cannot be accessed until explicitly imported
    from scipy import ndimage
except ImportError:
    scipy = None

from deepcell.image_generators import _transform_masks


class CroppingIterator(SemanticIterator):
    """Iterator yielding data from Numpy arrays (X and y).

    Args:
        train_dict (dict): Consists of numpy arrays for ``X`` and ``y``.
        image_data_generator (ImageDataGenerator): For random transformations
            and normalization.
        batch_size (int): Size of a batch.
        min_objects (int): Images with fewer than ``min_objects`` are ignored.
        shuffle (bool): Whether to shuffle the data between epochs.
        seed (int): Random seed for data shuffling.
        data_format (str): A string, one of ``channels_last`` (default)
            or ``channels_first``. The ordering of the dimensions in the
            inputs. ``channels_last`` corresponds to inputs with shape
            ``(batch, height, width, channels)`` while ``channels_first``
            corresponds to inputs with shape
            ``(batch, channels, height, width)``.
        save_to_dir (str): Optional directory where to save the pictures
            being yielded, in a viewable format. This is useful
            for visualizing the random transformations being
            applied, for debugging purposes.
        save_prefix (str): Prefix to use for saving sample
            images (if ``save_to_dir`` is set).
        save_format (str): Format to use for saving sample images
            (if ``save_to_dir`` is set).
        crop_size (tuple): Optional parameter specifying size of crop to take from image
    """
    def __init__(self,
                 train_dict,
                 image_data_generator,
                 batch_size=1,
                 shuffle=False,
                 transforms=['outer-distance'],
                 transforms_kwargs={},
                 seed=None,
                 min_objects=3,
                 data_format='channels_last',
                 save_to_dir=None,
                 save_prefix='',
                 save_format='png',
                 crop_size=None):

        super(CroppingIterator, self).__init__(
            train_dict=train_dict,
            image_data_generator=image_data_generator,
            batch_size=batch_size,
            shuffle=shuffle,
            transforms=transforms,
            transforms_kwargs=transforms_kwargs,
            seed=seed,
            min_objects=min_objects,
            data_format=data_format,
            save_to_dir=save_to_dir,
            save_prefix=save_prefix,
            save_format=save_format,
        )

        # set output size of image based on crop_size
        if crop_size is not None:
            output_size = crop_size

        else:
            output_size = self.x.shape[1:3] if self.channel_axis == 3 else self.x.shape[2:4]

        self.output_size = output_size

    def _get_batches_of_transformed_samples(self, index_array):
        # set output size based on output shape and # of channels
        if self.channel_axis == 3:
            x_shape = tuple([len(index_array)] + list(self.output_size) + [self.x.shape[3]])
        else:
            x_shape = tuple([len(index_array)] + [self.x.shape[1]] + list(self.output_size))

        batch_x = np.zeros(x_shape, dtype=self.x.dtype)
        batch_y = []

        for i, j in enumerate(index_array):
            x = self.x[j]

            # _transform_labels expects batch dimension
            y_semantic_list = self._transform_labels(self.y[j:j + 1])

            # initialize batch_y
            if len(batch_y) == 0:
                for ys in y_semantic_list:
                    if self.data_format == 'channels_first':
                        shape = tuple([len(index_array), ys.shape[1]] + list(self.output_size))
                    else:
                        shape = tuple([len(index_array)] + list(self.output_size) + [ys.shape[-1]])
                    batch_y.append(np.zeros(shape, dtype=ys.dtype))

            # random_transform does not expect batch dimension
            y_semantic_list = [ys[0] for ys in y_semantic_list]

            # Apply transformation
            x, y_semantic_list = self.image_data_generator.random_transform(
                x, y_semantic_list)

            x = self.image_data_generator.standardize(x)

            batch_x[i] = x

            for k, ys in enumerate(y_semantic_list):
                batch_y[k][i] = ys

        if self.save_to_dir:
            for i, j in enumerate(index_array):
                if self.data_format == 'channels_first':
                    img_x = np.expand_dims(batch_x[i, 0, ...], 0)
                else:
                    img_x = np.expand_dims(batch_x[i, ..., 0], -1)
                img = array_to_img(img_x, self.data_format, scale=True)
                fname = '{prefix}_{index}_{hash}.{format}'.format(
                    prefix=self.save_prefix,
                    index=j,
                    hash=np.random.randint(1e4),
                    format=self.save_format)
                img.save(os.path.join(self.save_to_dir, fname))

                if self.y is not None:
                    # Save argmax of y batch
                    for k, y_sem in enumerate(batch_y):
                        if y_sem[i].shape[self.channel_axis - 1] == 1:
                            img_y = y_sem[i]
                        else:
                            img_y = np.argmax(y_sem[i],
                                              axis=self.channel_axis - 1)
                            img_y = np.expand_dims(img_y,
                                                   axis=self.channel_axis - 1)
                        img = array_to_img(img_y, self.data_format, scale=True)
                        fname = 'y_{sem}_{prefix}_{index}_{hash}.{format}'.format(
                            sem=k,
                            prefix=self.save_prefix,
                            index=j,
                            hash=np.random.randint(1e4),
                            format=self.save_format)
                        img.save(os.path.join(self.save_to_dir, fname))

        return batch_x, batch_y


class CroppingDataGenerator(SemanticDataGenerator):
    """Generates batches of tensor image data with real-time data augmentation.
    The data will be looped over (in batches).

    Args:
        ffeaturewise_center (bool): Set input mean to 0 over the dataset,
            feature-wise.
        samplewise_center (bool): Set each sample mean to 0.
        featurewise_std_normalization (bool): Divide inputs by std
            of the dataset, feature-wise.
        samplewise_std_normalization (bool): Divide each input by its std.
        zca_epsilon (float): Epsilon for ZCA whitening. Default is 1e-6.
        zca_whitening (bool): Apply ZCA whitening.
        rotation_range (int): Degree range for random rotations.
        width_shift_range (float): 1-D array-like or int

            - float: fraction of total width, if < 1, or pixels if >= 1.
            - 1-D array-like: random elements from the array.
            - int: integer number of pixels from interval
              ``(-width_shift_range, +width_shift_range)``
            - With ``width_shift_range=2`` possible values are integers
              ``[-1, 0, +1]``, same as with ``width_shift_range=[-1, 0, +1]``,
              while with ``width_shift_range=1.0`` possible values are floats
              in the interval [-1.0, +1.0).

        height_shift_range: Float, 1-D array-like or int

            - float: fraction of total height, if < 1, or pixels if >= 1.
            - 1-D array-like: random elements from the array.
            - int: integer number of pixels from interval
              ``(-height_shift_range, +height_shift_range)``
            - With ``height_shift_range=2`` possible values
              are integers ``[-1, 0, +1]``,
              same as with ``height_shift_range=[-1, 0, +1]``,
              while with ``height_shift_range=1.0`` possible values are floats
              in the interval [-1.0, +1.0).

        shear_range (float): Shear Intensity
            (Shear angle in counter-clockwise direction in degrees)
        zoom_range (float): float or [lower, upper], Range for random zoom.
            If a float, ``[lower, upper] = [1-zoom_range, 1+zoom_range]``.
        channel_shift_range (float): range for random channel shifts.
        fill_mode (str): One of {"constant", "nearest", "reflect" or "wrap"}.

            Default is 'nearest'. Points outside the boundaries of the input
            are filled according to the given mode:

                - 'constant': kkkkkkkk|abcd|kkkkkkkk (cval=k)
                - 'nearest':  aaaaaaaa|abcd|dddddddd
                - 'reflect':  abcddcba|abcd|dcbaabcd
                - 'wrap':  abcdabcd|abcd|abcdabcd

        cval (float): Value used for points outside the boundaries
            when ``fill_mode = "constant"``.
        horizontal_flip (bool): Randomly flip inputs horizontally.
        vertical_flip (bool): Randomly flip inputs vertically.
        rescale: rescaling factor. Defaults to None. If None or 0, no rescaling
            is applied, otherwise we multiply the data by the value provided
            (before applying any other transformation).
        preprocessing_function: function that will be implied on each input.
            The function will run after the image is resized and augmented.
            The function should take one argument:
            one image (Numpy tensor with rank 3),
            and should output a Numpy tensor with the same shape.
        data_format (str): A string, one of ``channels_last`` (default)
            or ``channels_first``. The ordering of the dimensions in the
            inputs. ``channels_last`` corresponds to inputs with shape
            ``(batch, height, width, channels)`` while ``channels_first``
            corresponds to inputs with shape
            ``(batch, channels, height, width)``.
        validation_split (float): Fraction of images reserved for validation
            (strictly between 0 and 1).
    """

    def __init__(self,
                 featurewise_center=False,
                 samplewise_center=False,
                 featurewise_std_normalization=False,
                 samplewise_std_normalization=False,
                 zca_whitening=False,
                 zca_epsilon=1e-6,
                 rotation_range=0,
                 width_shift_range=0.,
                 height_shift_range=0.,
                 brightness_range=None,
                 shear_range=0.,
                 zoom_range=0.,
                 channel_shift_range=0.,
                 fill_mode='nearest',
                 cval=0.,
                 horizontal_flip=False,
                 vertical_flip=False,
                 rescale=None,
                 preprocessing_function=None,
                 data_format='channels_last',
                 validation_split=0.0,
                 interpolation_order=1,
                 crop_size=None,
                 dtype='float32'):

        super(CroppingDataGenerator, self).__init__(
            featurewise_center=featurewise_center,
            samplewise_center=samplewise_center,
            featurewise_std_normalization=featurewise_std_normalization,
            samplewise_std_normalization=samplewise_std_normalization,
            zca_whitening=zca_whitening,
            zca_epsilon=zca_epsilon,
            rotation_range=rotation_range,
            width_shift_range=width_shift_range,
            height_shift_range=height_shift_range,
            brightness_range=brightness_range,
            shear_range=shear_range,
            zoom_range=zoom_range,
            channel_shift_range=channel_shift_range,
            fill_mode=fill_mode,
            cval=cval,
            horizontal_flip=horizontal_flip,
            vertical_flip=vertical_flip,
            rescale=rescale,
            preprocessing_function=preprocessing_function,
            data_format=data_format,
            validation_split=validation_split,
            dtype=dtype)

        if crop_size is not None:
            if not isinstance(crop_size, (tuple, list)):
                raise ValueError("Crop size must be a list or tuple of row/col dimensions")

        self.crop_size = crop_size

        # tensorflow does not initialize interpolation_order, so we'll do it here
        self.interpolation_order = interpolation_order

    def flow(self,
             train_dict,
             batch_size=1,
             transforms=['outer-distance'],
             transforms_kwargs={},
             min_objects=3,
             shuffle=True,
             seed=None,
             save_to_dir=None,
             save_prefix='',
             save_format='png'):
        """Generates batches of augmented/normalized data with given arrays.

        Args:
            train_dict (dict): Consists of numpy arrays for ``X`` and ``y``.
            batch_size (int): Size of a batch. Defaults to 1.
            shuffle (bool): Whether to shuffle the data between epochs.
                Defaults to ``True``.
            seed (int): Random seed for data shuffling.
            min_objects (int): Minumum number of objects allowed per image
            save_to_dir (str): Optional directory where to save the pictures
                being yielded, in a viewable format. This is useful
                for visualizing the random transformations being
                applied, for debugging purposes.
            save_prefix (str): Prefix to use for saving sample
                images (if ``save_to_dir`` is set).
            save_format (str): Format to use for saving sample images
                (if ``save_to_dir`` is set).

        Returns:
            CroppingIterator: An ``Iterator`` yielding tuples of ``(x, y)``,
            where ``x`` is a numpy array of image data and ``y`` is list of
            numpy arrays of transformed masks of the same shape.
        """
        return CroppingIterator(
            train_dict,
            self,
            batch_size=batch_size,
            transforms=transforms,
            transforms_kwargs=transforms_kwargs,
            shuffle=shuffle,
            min_objects=min_objects,
            seed=seed,
            data_format=self.data_format,
            save_to_dir=save_to_dir,
            save_prefix=save_prefix,
            save_format=save_format,
            crop_size=self.crop_size)

    def get_random_transform(self, img_shape, seed=None):
        transform_parameters = super(CroppingDataGenerator, self).get_random_transform(
            img_shape=img_shape, seed=seed)

        crop_indices = None
        if self.crop_size is not None:

            img_dims = img_shape[1:] if self.channel_axis == 1 else img_shape[:2]
            if img_dims == self.crop_size:
                # don't need to crop
                pass
            elif img_dims[0] == self.crop_size[0] or img_dims[1] == self.crop_size[1]:
                raise ValueError('crop_size must be a subset of both axes or exactly '
                                 ' equal to image dims')
            elif img_dims[0] < self.crop_size[0] or img_dims[1] < self.crop_size[1]:
                raise ValueError('Crop dimensions must be smaller than image dimensions')
            else:
                row_start = np.random.randint(0, img_dims[0] - self.crop_size[0])
                col_start = np.random.randint(0, img_dims[1] - self.crop_size[1])
                crop_indices = ([row_start, row_start + self.crop_size[0]],
                                [col_start, col_start + self.crop_size[1]])

        transform_parameters['crop_indices'] = crop_indices

        return transform_parameters

    def apply_transform(self, x, transform_parameters):

        if transform_parameters['crop_indices'] is not None:
            row_indices, col_indices = transform_parameters['crop_indices']
            if self.channel_axis == 1:
                x = x[:, row_indices[0]:row_indices[1], col_indices[0]:col_indices[1]]
            else:
                x = x[row_indices[0]:row_indices[1], col_indices[0]:col_indices[1], :]

        x = super(CroppingDataGenerator, self).apply_transform(
            x=x, transform_parameters=transform_parameters)
        return x

    def fit(self, x, augment=False, rounds=1, seed=None):
        """Fits the data generator to some sample data.
        This computes the internal data stats related to the
        data-dependent transformations, based on an array of sample data.
        Only required if `featurewise_center` or
        `featurewise_std_normalization` or `zca_whitening` are set to True.
        When `rescale` is set to a value, rescaling is applied to
        sample data before computing the internal data stats.

        Args:
            x: Sample data. Should have rank 4.
             In case of grayscale data,
             the channels axis should have value 1, in case
             of RGB data, it should have value 3, and in case
             of RGBA data, it should have value 4.
            augment: Boolean (default: False).
                Whether to fit on randomly augmented samples.
            rounds: Int (default: 1).
                If using data augmentation (`augment=True`),
                this is how many augmentation passes over the data to use.
            seed: Int (default: None). Random seed.
       """
        x = np.asarray(x, dtype=self.dtype)
        if x.ndim != 4:
            raise ValueError('Input to `.fit()` should have rank 4. '
                             'Got array with shape: ' + str(x.shape))
        if x.shape[self.channel_axis] not in {1, 3, 4}:
            warnings.warn(
                'Expected input to be images (as Numpy array) '
                'following the data format convention "' +
                self.data_format + '" (channels on axis ' +
                str(self.channel_axis) + '), i.e. expected '
                                         'either 1, 3 or 4 channels on axis ' +
                str(self.channel_axis) + '. '
                                         'However, it was passed an array with shape ' +
                str(x.shape) + ' (' + str(x.shape[self.channel_axis]) +
                ' channels).')

        if seed is not None:
            np.random.seed(seed)

        x = np.copy(x)
        if self.rescale:
            x *= self.rescale

        if augment:
            # adjust output shape to account for cropping in generator
            if self.crop_size is not None:
                if self.channel_axis == 1:
                    x_crop_shape = [x.shape[1]] + list(self.crop_size)
                else:
                    x_crop_shape = list(self.crop_size) + [x.shape[3]]

                ax = np.zeros(
                    tuple([rounds * x.shape[0]] + x_crop_shape),
                    dtype=self.dtype)
            else:
                ax = np.zeros(
                    tuple([rounds * x.shape[0]] + list(x.shape)[1:]),
                    dtype=self.dtype)

            for r in range(rounds):
                for i in range(x.shape[0]):
                    ax[i + r * x.shape[0]] = self.random_transform(x[i])
            x = ax

        if self.featurewise_center:
            self.mean = np.mean(x, axis=(0, self.row_axis, self.col_axis))
            broadcast_shape = [1, 1, 1]
            broadcast_shape[self.channel_axis - 1] = x.shape[self.channel_axis]
            self.mean = np.reshape(self.mean, broadcast_shape)
            x -= self.mean

        if self.featurewise_std_normalization:
            self.std = np.std(x, axis=(0, self.row_axis, self.col_axis))
            broadcast_shape = [1, 1, 1]
            broadcast_shape[self.channel_axis - 1] = x.shape[self.channel_axis]
            self.std = np.reshape(self.std, broadcast_shape)
            x /= (self.std + 1e-6)

        if self.zca_whitening:
            if scipy is None:
                raise ImportError('Using zca_whitening requires SciPy. '
                                  'Install SciPy.')
            flat_x = np.reshape(
                x, (x.shape[0], x.shape[1] * x.shape[2] * x.shape[3]))
            sigma = np.dot(flat_x.T, flat_x) / flat_x.shape[0]
            u, s, _ = linalg.svd(sigma)
            s_inv = 1. / np.sqrt(s[np.newaxis] + self.zca_epsilon)
            self.principal_components = (u * s_inv).dot(u.T)



# Copyright 2016-2021 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tf/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for backbone_utils"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

from absl.testing import parameterized

from tensorflow.python.framework import test_util as tf_test_util
from tensorflow.python.platform import test

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Input
from tensorflow.keras.models import Model
from tensorflow.python.keras import keras_parameterized

from deepcell.utils import backbone_utils


class TestBackboneUtils(keras_parameterized.TestCase):

    @keras_parameterized.run_with_all_model_types
    @keras_parameterized.run_all_keras_modes
    @parameterized.named_parameters(
        *tf_test_util.generate_combinations_with_testcase_name(
            data_format=[
                # 'channels_first',
                'channels_last']))
    def test_get_featurenet_backbone(self, data_format):
        backbone = 'featurenet'
        input_shape = (256, 256, 3)
        inputs = Input(shape=input_shape)
        with self.cached_session():
            K.set_image_data_format(data_format)
            model, output_dict = backbone_utils.get_backbone(
                backbone, inputs, return_dict=True)
            assert isinstance(output_dict, dict)
            assert all(k.startswith('C') for k in output_dict)
            assert isinstance(model, Model)

            # No imagenet weights for featurenet backbone
            with self.assertRaises(ValueError):
                backbone_utils.get_backbone(backbone, inputs, use_imagenet=True)

    # @keras_parameterized.run_all_keras_modes
    @parameterized.named_parameters(
        *tf_test_util.generate_combinations_with_testcase_name(
            data_format=[
                # 'channels_first',
                'channels_last']))
    def test_get_featurenet3d_backbone(self, data_format):
        backbone = 'featurenet3d'
        input_shape = (40, 256, 256, 3)
        inputs = Input(shape=input_shape)
        with self.cached_session():
            K.set_image_data_format(data_format)
            model, output_dict = backbone_utils.get_backbone(
                backbone, inputs, return_dict=True)
            assert isinstance(output_dict, dict)
            assert all(k.startswith('C') for k in output_dict)
            assert isinstance(model, Model)

            # No imagenet weights for featurenet backbone
            with self.assertRaises(ValueError):
                backbone_utils.get_backbone(backbone, inputs, use_imagenet=True)

    # @keras_parameterized.run_with_all_model_types
    # @keras_parameterized.run_all_keras_modes
    @parameterized.named_parameters(
        *tf_test_util.generate_combinations_with_testcase_name(
            backbone=[
                'resnet50',
                'resnet101',
                'resnet152',
                'resnet50v2',
                'resnet101v2',
                'resnet152v2',
                # 'resnext50',
                # 'resnext101',
                'vgg16',
                'vgg19',
                'densenet121',
                'densenet169',
                'densenet201',
                'mobilenet',
                'mobilenetv2',
                'efficientnetb0',
                'efficientnetb1',
                'efficientnetb2',
                'efficientnetb3',
                'efficientnetb4',
                'efficientnetb5',
                'efficientnetb6',
                'efficientnetb7',
                'nasnet_large',
                'nasnet_mobile']))
    def test_get_backbone(self, backbone):
        with self.cached_session():
            K.set_image_data_format('channels_last')
            inputs = Input(shape=(256, 256, 3))
            model, output_dict = backbone_utils.get_backbone(
                backbone, inputs, return_dict=True)
            assert isinstance(output_dict, dict)
            assert all(k.startswith('C') for k in output_dict)
            assert isinstance(model, Model)

    def test_invalid_backbone(self):
        inputs = Input(shape=(4, 2, 3))
        with self.assertRaises(ValueError):
            backbone_utils.get_backbone('bad', inputs, return_dict=True)


if __name__ == '__main__':
    test.main()



# Copyright 2016-2021 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tf/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for tracking functions"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import parameterized

import numpy as np
import tensorflow as tf

from tensorflow.python.platform import test

from deepcell.data import tracking


@parameterized.named_parameters([
    {
        'testcase_name': 'test_data_01',
        'time': 30,
        'max_cells': 25,
        'crop_size': 32,
        'batch_size': 32,
        'seed': None,
        'track_length': 8,
        'val_size': 0.15,
        'test_size': 0.15
    }, {
        'testcase_name': 'test_data_02',
        'time': 40,
        'max_cells': 29,
        'crop_size': 32,
        'batch_size': 24,
        'seed': 2,
        'track_length': 7,
        'val_size': 0.2,
        'test_size': 0
    }
])
class TrackingTests(test.TestCase, parameterized.TestCase):

    def test_temporal_slice(self, time, max_cells, crop_size,
                            batch_size, seed, track_length,
                            val_size, test_size):
        X, y = self.create_test_data(time, max_cells, crop_size)
        sliced_X, sliced_y = tracking.temporal_slice(X, y, track_length)

        # Test temporal dimension correct index and value
        for key in sliced_X:
            self.assertEqual(sliced_X[key].shape[0], track_length)
        for key in sliced_y:
            self.assertEqual(sliced_y[key].shape[0], track_length - 1)

    def test_random_rotate(self, time, max_cells, crop_size,
                           batch_size, seed, track_length,
                           val_size, test_size):
        X, y = self.create_test_data(time, max_cells, crop_size)

        # Get appearance and centroid tensors
        X_apps = X['appearances']
        X_cents = X['centroids']

        # Assert rotating does not mutate the shape of tensors
        rotated_X, y = tracking.random_rotate(X, y, 180)
        self.assertEqual(rotated_X['appearances'].shape, X_apps.shape)
        self.assertEqual(rotated_X['centroids'].shape, X_cents.shape)

        # Assert rotating by 0 does not change tensors
        X_apps = rotated_X['appearances']
        X_cents = rotated_X['centroids']

        output_X, y = tracking.random_rotate(rotated_X, y, 0)

        self.assertAllEqual(output_X['appearances'], X_apps)
        self.assertAllEqual(output_X['centroids'], X_cents)

    def test_random_translate(self, time, max_cells, crop_size,
                              batch_size, seed, track_length,
                              val_size, test_size):
        X, y = self.create_test_data(time, max_cells, crop_size)
        X_cents = X['centroids']
        translated_X, y = tracking.random_translate(X, y)

        # Get difference from translating
        diff = X_cents - translated_X['centroids']

        # Create array filled with same x and y values
        x0 = np.empty((time, max_cells, 1))
        y0 = np.empty((time, max_cells, 1))
        x0.fill(diff[0, 0, 0])
        y0.fill(diff[0, 0, 1])
        r0 = tf.concat([x0, y0], 2)

        # Assert that centroids are translated by the same amount
        self.assertAllEqual(diff, r0)

        # Assert range of 0 does not translate data
        X_cents = translated_X['centroids']
        output_X, y = tracking.random_translate(X, y, range=0)
        self.assertAllEqual(output_X['centroids'], X_cents)

    def test_prepare_dataset(self, time, max_cells, crop_size,
                             batch_size, seed, track_length,
                             val_size, test_size):
        # Create track_info and prepare the dataset
        track_info = self.create_track_info(226, time, max_cells, crop_size)
        train_data, val_data, test_data = tracking.prepare_dataset(track_info,
                                                                   rotation_range=180,
                                                                   batch_size=batch_size,
                                                                   seed=seed,
                                                                   track_length=track_length,
                                                                   val_size=val_size,
                                                                   test_size=test_size)
        # Test data correctly batched
        for X, y in train_data.take(1):
            self.assertEqual((X['appearances'].shape)[0], batch_size)
        for X, y in val_data.take(1):
            self.assertEqual((X['appearances'].shape)[0], batch_size)
        if test_size != 0:
            for X, y in test_data.take(1):
                self.assertEqual((X['appearances'].shape)[0], batch_size)

    def create_test_data(self, time, max_cells, crop_size):
        # Create dictionaries of feature data labels with correct dimensions
        X = {}
        y = {}
        X['appearances'] = tf.random.uniform([time, max_cells, crop_size, crop_size, 1], 0, 1)
        X['centroids'] = tf.random.uniform([time, max_cells, 2], 0, 512)
        X['morphologies'] = tf.random.uniform([time, max_cells, 3], 0, 1)
        X['adj_matrices'] = tf.random.uniform([time, max_cells, max_cells], 0, 1)
        y['temporal_adj_matrices'] = tf.random.uniform([time - 1, max_cells, max_cells, 3], 0, 1)
        return X, y

    def create_track_info(self, n_batches, time, max_cells, crop_size):
        # Create track_info input (dictionary of all input and output features) with correct
        # dimensions for prepare_dataset function.
        track_info = {}
        track_info['appearances'] = tf.random.uniform([n_batches, time, max_cells,
                                                       crop_size, crop_size, 1], 0, 1)
        track_info['centroids'] = tf.random.uniform([n_batches, time, max_cells, 2], 0, 512)
        track_info['morphologies'] = tf.random.uniform([n_batches, time, max_cells, 3], 0, 1)
        track_info['adj_matrices'] = tf.random.uniform([n_batches, time, max_cells, max_cells],
                                                       0, 1)
        track_info['norm_adj_matrices'] = tf.random.uniform([n_batches, time, max_cells,
                                                            max_cells], 0, 1)
        track_info['temporal_adj_matrices'] = tf.random.uniform([n_batches, time - 1, max_cells,
                                                                max_cells, 3], 0, 1)
        return track_info


if __name__ == '__main__':
    test.main()



# Copyright 2016-2021 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tf/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Custom loss functions for DeepCell"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import tensorflow as tf
from tensorflow.keras import backend as K


def categorical_crossentropy(y_true, y_pred, class_weights=None, axis=None, from_logits=False):
    """Categorical crossentropy between an output tensor and a target tensor.

    Args:
        y_true: A tensor of the same shape as output.
        y_pred: A tensor resulting from a softmax
            (unless ``from_logits`` is ``True``, in which
            case ``y_pred`` is expected to be the logits).
        from_logits: Boolean, whether ``y_pred`` is the
            result of a softmax, or is a tensor of logits.

    Returns:
        tensor: Output tensor.
    """
    # Note: tf.nn.softmax_cross_entropy_with_logits
    # expects logits, Keras expects probabilities.
    y_pred = tf.convert_to_tensor(y_pred)
    y_true = K.cast(y_true, y_pred.dtype)
    if axis is None:
        axis = 1 if K.image_data_format() == 'channels_first' else K.ndim(y_pred) - 1
    if not from_logits:
        # scale preds so that the class probas of each sample sum to 1
        y_pred = y_pred / K.sum(y_pred, axis=axis, keepdims=True)
        # manual computation of crossentropy
        _epsilon = tf.convert_to_tensor(K.epsilon(), y_pred.dtype.base_dtype)
        y_pred = tf.clip_by_value(y_pred, _epsilon, 1. - _epsilon)
        if class_weights is None:
            return - K.sum(y_true * K.log(y_pred), axis=axis)
        return - K.sum((y_true * K.log(y_pred) * class_weights), axis=axis)
    return tf.nn.softmax_cross_entropy_with_logits(labels=y_true, logits=y_pred)


def weighted_categorical_crossentropy(y_true, y_pred,
                                      n_classes=3, axis=None,
                                      from_logits=False):
    """Categorical crossentropy between an output tensor and a target tensor.
    Automatically computes the class weights from the target image and uses
    them to weight the cross entropy

    Args:
        y_true: A tensor of the same shape as ``y_pred``.
        y_pred: A tensor resulting from a softmax
            (unless ``from_logits`` is ``True``, in which
            case ``y_pred`` is expected to be the logits).
        from_logits: Boolean, whether ``y_pred`` is the
            result of a softmax, or is a tensor of logits.

    Returns:
        tensor: Output tensor.
    """
    if from_logits:
        raise Exception('weighted_categorical_crossentropy cannot take logits')
    y_pred = tf.convert_to_tensor(y_pred)
    y_true = K.cast(y_true, y_pred.dtype)
    n_classes = K.cast(n_classes, y_pred.dtype)
    if axis is None:
        axis = 1 if K.image_data_format() == 'channels_first' else K.ndim(y_pred) - 1
    reduce_axis = [x for x in list(range(K.ndim(y_pred))) if x != axis]
    # scale preds so that the class probas of each sample sum to 1
    y_pred = y_pred / K.sum(y_pred, axis=axis, keepdims=True)
    # manual computation of crossentropy
    _epsilon = tf.convert_to_tensor(K.epsilon(), y_pred.dtype.base_dtype)
    y_pred = tf.clip_by_value(y_pred, _epsilon, 1. - _epsilon)
    total_sum = K.sum(y_true)
    class_sum = K.sum(y_true, axis=reduce_axis, keepdims=True)
    class_weights = 1.0 / n_classes * tf.divide(total_sum, class_sum + 1.)
    return - K.sum((y_true * K.log(y_pred) * class_weights), axis=axis)


def sample_categorical_crossentropy(y_true,
                                    y_pred,
                                    class_weights=None,
                                    axis=None,
                                    from_logits=False):
    """Categorical crossentropy between an output tensor and a target tensor.
    Only the sampled pixels defined by y_true = 1 are used to compute the
    cross entropy.

    Args:
        y_true: A tensor of the same shape as ``y_pred``.
        y_pred: A tensor resulting from a softmax
            (unless ``from_logits`` is ``True``, in which
            case ``y_pred`` is expected to be the logits).
        from_logits: Boolean, whether ``y_pred`` is the
            result of a softmax, or is a tensor of logits.

    Returns:
        tensor: Output tensor.
    """
    # Note: tf.nn.softmax_cross_entropy_with_logits
    # expects logits, Keras expects probabilities.
    y_pred = tf.convert_to_tensor(y_pred)
    y_true = K.cast(y_true, y_pred.dtype)
    if axis is None:
        axis = 1 if K.image_data_format() == 'channels_first' else K.ndim(y_pred) - 1
    if not from_logits:
        # scale preds so that the class probabilities of each sample sum to 1
        y_pred = y_pred / K.sum(y_pred, axis=axis, keepdims=True)

        # Multiply with mask so that only the sampled pixels are used
        y_pred = y_pred * y_true

        # manual computation of crossentropy
        _epsilon = tf.convert_to_tensor(K.epsilon(), y_pred.dtype.base_dtype)
        y_pred = tf.clip_by_value(y_pred, _epsilon, 1. - _epsilon)
        if class_weights is None:
            return - K.sum(y_true * K.log(y_pred), axis=axis)
        return - K.sum((y_true * K.log(y_pred) * class_weights), axis=axis)
    return tf.nn.softmax_cross_entropy_with_logits(labels=y_true, logits=y_pred)


def dice_loss(y_true, y_pred, smooth=1):
    """Dice coefficient loss between an output tensor and a target tensor.

    Args:
        y_true: A tensor of the same shape as ``y_pred``.
        y_pred: A tensor resulting from a softmax

    Returns:
        tensor: Output tensor.
    """
    y_pred = tf.convert_to_tensor(y_pred)
    y_true = K.cast(y_true, y_pred.dtype)
    y_true_f = K.flatten(y_true)
    y_pred_f = K.flatten(y_pred)
    intersection = K.sum(y_true_f * y_pred_f)
    return - ((2. * intersection + smooth) / (K.sum(y_true_f) + K.sum(y_pred_f) + smooth))


def discriminative_instance_loss(y_true, y_pred,
                                 delta_v=0.5,
                                 delta_d=1.5,
                                 gamma=1e-3):
    """Discriminative loss between an output tensor and a target tensor.

    Args:
        y_true: A tensor of the same shape as ``y_pred``.
        y_pred: A tensor of the vector embedding

    Returns:
        tensor: Output tensor.
    """
    y_pred = tf.convert_to_tensor(y_pred)
    y_true = K.cast(y_true, y_pred.dtype)

    def temp_norm(ten, axis=None):
        if axis is None:
            axis = 1 if K.image_data_format() == 'channels_first' else K.ndim(ten) - 1
        return K.sqrt(K.epsilon() + K.sum(K.square(ten), axis=axis))

    rank = K.ndim(y_pred)
    channel_axis = 1 if K.image_data_format() == 'channels_first' else rank - 1
    axes = [x for x in list(range(rank)) if x != channel_axis]

    # Compute variance loss
    cells_summed = tf.tensordot(y_true, y_pred, axes=[axes, axes])
    nonzeros = tf.math.count_nonzero(y_true, axis=axes)
    n_pixels = K.cast(nonzeros, dtype=y_pred.dtype) + K.epsilon()
    n_pixels_expand = K.expand_dims(n_pixels, axis=1) + K.epsilon()
    mu = tf.divide(cells_summed, n_pixels_expand)

    delta_v = K.constant(delta_v, dtype=y_pred.dtype)
    mu_tensor = tf.tensordot(y_true, mu, axes=[[channel_axis], [0]])
    L_var_1 = y_pred - mu_tensor
    L_var_2 = K.square(K.relu(temp_norm(L_var_1) - delta_v))
    L_var_3 = tf.tensordot(L_var_2, y_true, axes=[axes, axes])
    L_var_4 = tf.divide(L_var_3, n_pixels)
    L_var = K.mean(L_var_4)

    # Compute distance loss
    mu_a = K.expand_dims(mu, axis=0)
    mu_b = K.expand_dims(mu, axis=1)

    diff_matrix = tf.subtract(mu_b, mu_a)
    L_dist_1 = temp_norm(diff_matrix)
    L_dist_2 = K.square(K.relu(K.constant(2 * delta_d, dtype=y_pred.dtype) - L_dist_1))
    diag = K.constant(0, dtype=y_pred.dtype) * tf.linalg.diag_part(L_dist_2)
    L_dist_3 = tf.linalg.set_diag(L_dist_2, diag)
    L_dist = K.mean(L_dist_3)

    # Compute regularization loss
    L_reg = gamma * temp_norm(mu)
    L = L_var + L_dist + K.mean(L_reg)

    return L


def weighted_focal_loss(y_true, y_pred, n_classes=3, gamma=2., axis=None, from_logits=False):
    """Focal loss between an output tensor and a target tensor.
    Automatically computes the class weights from the target image and uses
    them to weight the cross entropy

    Args:
        y_true: A tensor of the same shape as ``y_pred``.
        y_pred: A tensor resulting from a softmax
            (unless ``from_logits`` is ``True``, in which
            case ``y_pred`` is expected to be the logits).
        from_logits: Boolean, whether ``y_pred`` is the
            result of a softmax, or is a tensor of logits.

    Returns:
        tensor: Output tensor.
    """
    if from_logits:
        raise Exception('weighted_focal_loss cannot take logits')
    y_pred = tf.convert_to_tensor(y_pred)
    y_true = K.cast(y_true, y_pred.dtype)
    n_classes = K.cast(n_classes, y_pred.dtype)
    if axis is None:
        axis = 1 if K.image_data_format() == 'channels_first' else K.ndim(y_pred) - 1
    reduce_axis = [x for x in list(range(K.ndim(y_pred))) if x != axis]
    # scale preds so that the class probas of each sample sum to 1
    y_pred = y_pred / K.sum(y_pred, axis=axis, keepdims=True)
    # manual computation of crossentropy
    _epsilon = tf.convert_to_tensor(K.epsilon(), y_pred.dtype.base_dtype)
    y_pred = tf.clip_by_value(y_pred, _epsilon, 1. - _epsilon)
    total_sum = K.sum(y_true)
    class_sum = K.sum(y_true, axis=reduce_axis, keepdims=True)
    class_weights = 1.0 / n_classes * tf.divide(total_sum, class_sum + 1.)
    temp_loss = (K.pow(1. - y_pred, gamma) * K.log(y_pred) * class_weights)
    focal_loss = - K.sum(y_true * temp_loss, axis=axis)
    return focal_loss


def smooth_l1(y_true, y_pred, sigma=3.0, axis=None):
    """Compute the smooth L1 loss of ``y_pred`` w.r.t. ``y_true``.

    Args:
        y_true: Tensor from the generator of shape ``(B, N, 5)``.
            The last value for each box is the state of the anchor
            (ignore, negative, positive).
        y_pred: Tensor from the network of shape ``(B, N, 4)``.
        sigma: The point where the loss changes from L2 to L1.

    Returns:
        The smooth L1 loss of ``y_pred`` w.r.t. ``y_true``.
    """
    y_pred = tf.convert_to_tensor(y_pred)
    y_true = K.cast(y_true, y_pred.dtype)

    if axis is None:
        axis = 1 if K.image_data_format() == 'channels_first' else K.ndim(y_pred) - 1

    sigma_squared = sigma ** 2

    # compute smooth L1 loss
    # f(x) = 0.5 * (sigma * x)^2          if |x| < 1 / sigma / sigma
    #        |x| - 0.5 / sigma / sigma    otherwise
    regression_diff = K.abs(y_true - y_pred)  # |y - f(x)|

    regression_loss = tf.where(
        K.less(regression_diff, 1.0 / sigma_squared),
        0.5 * sigma_squared * K.pow(regression_diff, 2),
        regression_diff - 0.5 / sigma_squared)
    return K.sum(regression_loss, axis=axis)


def focal(y_true, y_pred, alpha=0.25, gamma=2.0, axis=None):
    """Compute the focal loss given the target tensor and the predicted tensor.

    As defined in https://arxiv.org/abs/1708.02002

    Args:
        y_true: Tensor of target data with shape ``(B, N, num_classes)``.
        y_pred: Tensor of predicted data with shape ``(B, N, num_classes)``.
        alpha: Scale the focal weight with ``alpha``.
        gamma: Take the power of the focal weight with ``gamma``.

    Returns:
        float: The focal loss of ``y_pred`` w.r.t. ``y_true``.
    """
    y_pred = tf.convert_to_tensor(y_pred)
    y_true = K.cast(y_true, y_pred.dtype)

    if axis is None:
        axis = 1 if K.image_data_format() == 'channels_first' else K.ndim(y_pred) - 1

    # compute the focal loss
    alpha_factor = K.ones_like(y_true) * alpha
    alpha_factor = tf.where(K.equal(y_true, 1), alpha_factor, 1 - alpha_factor)
    focal_weight = tf.where(K.equal(y_true, 1), 1 - y_pred, y_pred)
    focal_weight = alpha_factor * focal_weight ** gamma

    cls_loss = focal_weight * K.binary_crossentropy(y_true, y_pred)

    return K.sum(cls_loss, axis=axis)
